from .attributes import Attribute
from .constants import Utf8


# From: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html


##########################################################################
# 4.6. Methods
##########################################################################

# Each method, including each instance initialization method (§2.9) and the
# class or interface initialization method (§2.9), is described by a method_info
# structure. No two methods in one class file may have the same name and
# descriptor (§4.3.3).

class Method:
    # u2 access_flags;
    # u2 name_index;
    # u2 descriptor_index;
    # u2 attributes_count;
    # attribute_info attributes[attributes_count];

    # Table 4.5. Method access and property flags

    ACC_PUBLIC = 0x0001         # Declared public; may be accessed from outside its package.
    ACC_PRIVATE = 0x0002        # Declared private; accessible only within the defining class.
    ACC_PROTECTED = 0x0004      # Declared protected; may be accessed within subclasses.
    ACC_STATIC = 0x0008         # Declared static.
    ACC_FINAL = 0x0010          # Declared final; must not be overridden (§5.4.5).
    ACC_SYNCHRONIZED = 0x0020   # Declared synchronized; invocation is wrapped by a monitor use.
    ACC_BRIDGE = 0x0040         # A bridge method, generated by the compiler.
    ACC_VARARGS = 0x0080        # Declared with variable number of arguments.
    ACC_NATIVE = 0x0100         # Declared native; implemented in a language other than Java.
    ACC_ABSTRACT = 0x0400       # Declared abstract; no implementation is provided.
    ACC_STRICT = 0x0800         # Declared strictfp; floating-point mode is FP-strict.
    ACC_SYNTHETIC = 0x1000      # Declared synthetic; not present in the source code.

    def __init__(
                self, name, descriptor,
                public=True, private=False, protected=False, static=False,
                final=False, synchronized=False, bridge=False, varargs=False,
                native=False, abstract=False, strict=False, synthetic=False,
                attributes=None
            ):

        # The value of the name_index item must be a valid index into the
        # constant_pool table. The constant_pool entry at that index must be a
        # CONSTANT_Utf8_info (§4.4.7) structure representing either one of the
        # special method names (§2.9) <init> or <clinit>, or a valid unqualified
        # name (§4.2.2) denoting a method.
        self.name = Utf8(name)

        # The value of the descriptor_index item must be a valid index into the
        # constant_pool table. The constant_pool entry at that index must be a
        # CONSTANT_Utf8_info (§4.4.7) structure representing a valid method
        # descriptor (§4.3.3).

        # A future edition of this specification may require that the last
        # parameter descriptor of the method descriptor is an array type if the
        # ACC_VARARGS flag is set in the access_flags item.
        self.descriptor = Utf8(descriptor)

        # The ACC_VARARGS flag indicates that this method takes a variable
        # number of arguments at the source code level. A method declared to
        # take a variable number of arguments must be compiled with the
        # ACC_VARARGS flag set to 1. All other methods must be compiled with the
        # ACC_VARARGS flag set to 0.

        # The ACC_BRIDGE flag is used to indicate a bridge method generated by a
        # Java compiler.

        # A method may be marked with the ACC_SYNTHETIC flag to indicate that it
        # was generated by a compiler and does not appear in source code, unless
        # it is one of the methods named in §4.7.8.

        # Methods of classes may set any of the flags in Table 4.5. However, a
        # specific method of a class may have at most one of its ACC_PRIVATE,
        # ACC_PROTECTED and ACC_PUBLIC flags set (JLS §8.4.3). If a specific
        # method has its ACC_ABSTRACT flag set, it must not have any of its
        # ACC_FINAL, ACC_NATIVE, ACC_PRIVATE, ACC_STATIC, ACC_STRICT or
        # ACC_SYNCHRONIZED flags set (JLS §8.4.3.1, JLS §8.4.3.3, JLS §8.4.3.4).

        # All interface methods must have their ACC_ABSTRACT and ACC_PUBLIC
        # flags set; they may have their ACC_VARARGS, ACC_BRIDGE and
        # ACC_SYNTHETIC flags set and must not have any of the other flags in
        # Table 4.5 set (JLS §9.4).

        self.private = private
        if self.private:
            self.protected = False
            self.public = False
        else:
            self.protected = protected
            if self.protected:
                self.public = False
            else:
                self.public = True

        self.static = static
        self.final = final
        self.synchronized = synchronized
        self.bridge = bridge
        self.varargs = varargs
        self.native = native
        self.abstract = abstract
        self.strict = strict
        self.synthetic = synthetic

        # Each value of the attributes table must be an attribute structure
        # (§4.7). A method can have any number of optional attributes associated
        # with it.

        # The attributes defined by this specification as appearing in the
        # attributes table of a method_info structure are the Code (§4.7.3),
        # Exceptions (§4.7.5), Synthetic (§4.7.8), Signature (§4.7.9),
        # Deprecated (§4.7.15), RuntimeVisibleAnnotations (§4.7.16),
        # RuntimeInvisibleAnnotations (§4.7.17),
        # RuntimeVisibleParameterAnnotations (§4.7.18),
        # RuntimeInvisibleParameterAnnotations (§4.7.19), and AnnotationDefault
        # (§4.7.20) attributes.

        # A Java Virtual Machine implementation must recognize and correctly
        # read Code (§4.7.3) and Exceptions (§4.7.5) attributes found in the
        # attributes table of a method_info structure. If a Java Virtual Machine
        # implementation recognizes class files whose version number is 49.0 or
        # above, it must recognize and correctly read Signature (§4.7.9),
        # RuntimeVisibleAnnotations (§4.7.16), RuntimeInvisibleAnnotations
        # (§4.7.17), RuntimeVisibleParameterAnnotations (§4.7.18),
        # RuntimeInvisibleParameterAnnotations (§4.7.19) and AnnotationDefault
        # (§4.7.20) attributes found in the attributes table of a method_info
        # structure of a class file whose version number is 49.0 or above.

        # A Java Virtual Machine implementation is required to silently ignore
        # any or all attributes in the attributes table of a method_info
        # structure that it does not recognize. Attributes not defined in this
        # specification are not allowed to affect the semantics of the class
        # file, but only to provide additional descriptive information (§4.7.1).
        self.attributes = attributes if attributes else []

    def __repr__(self):
        return '<Method access:0x%04x name:%s, descriptor:%s>' % (self.access_flags, self.name, self.descriptor)

    @staticmethod
    def read(reader, dump=None):
        access_flags = reader.read_u2()

        name = reader.constant_pool[reader.read_u2()].bytes.decode('mutf-8')
        descriptor = reader.constant_pool[reader.read_u2()].bytes.decode('mutf-8')
        attributes_count = reader.read_u2()

        if dump is not None:
            reader.debug("    " * dump, 'Method %s %s' % (name, descriptor))

            access_description = ', '.join(f for f in [
                    flag if access_flags & mask else None
                    for flag, mask in [
                        ('public', Method.ACC_PUBLIC),
                        ('private', Method.ACC_PRIVATE),
                        ('protected', Method.ACC_PROTECTED),
                        ('static', Method.ACC_STATIC),
                        ('final', Method.ACC_FINAL),
                        ('synchronized', Method.ACC_SYNCHRONIZED),
                        ('bridge', Method.ACC_BRIDGE),
                        ('varargs', Method.ACC_VARARGS),
                        ('native', Method.ACC_NATIVE),
                        ('abstract', Method.ACC_ABSTRACT),
                        ('strict', Method.ACC_STRICT),
                        ('synthetic', Method.ACC_SYNTHETIC),
                    ]
                ] if f)
            reader.debug("    " * dump, '    Flags: 0x%04x%s' % (access_flags, ' (%s)') % (
                access_description if access_description else ''
            ))

            reader.debug("    " * dump, '    Attributes: (%s)' % attributes_count)

        attributes = []
        for i in range(0, attributes_count):
            attributes.append(
                Attribute.read(reader, dump=dump + 2 if dump is not None else dump)
            )

        return Method(
            name=name,
            descriptor=descriptor,
            public=bool(access_flags & Method.ACC_PUBLIC),
            private=bool(access_flags & Method.ACC_PRIVATE),
            protected=bool(access_flags & Method.ACC_PROTECTED),
            static=bool(access_flags & Method.ACC_STATIC),
            final=bool(access_flags & Method.ACC_FINAL),
            synchronized=bool(access_flags & Method.ACC_SYNCHRONIZED),
            bridge=bool(access_flags & Method.ACC_BRIDGE),
            varargs=bool(access_flags & Method.ACC_VARARGS),
            native=bool(access_flags & Method.ACC_NATIVE),
            abstract=bool(access_flags & Method.ACC_ABSTRACT),
            strict=bool(access_flags & Method.ACC_STRICT),
            synthetic=bool(access_flags & Method.ACC_SYNTHETIC),
            attributes=attributes,
        )

    def write(self, writer):
        writer.write_u2(self.access_flags)
        writer.write_u2(writer.constant_pool.index(self.name))
        writer.write_u2(writer.constant_pool.index(self.descriptor))
        writer.write_u2(self.attributes_count)

        for attribute in self.attributes:
            attribute.write(writer)

    def resolve(self, constant_pool):
        constant_pool.add(self.name)
        constant_pool.add(self.descriptor)

        for attribute in self.attributes:
            attribute.resolve(constant_pool)

    @property
    def attributes_count(self):
        return len(self.attributes)

    @property
    def access_flags(self):
        """A specific instance initialization method (§2.9) may have at most one
        of its ACC_PRIVATE, ACC_PROTECTED, and ACC_PUBLIC flags set, and may
        also have its ACC_STRICT, ACC_VARARGS and ACC_SYNTHETIC flags set, but
        must not have any of the other flags in Table 4.5 set.

        Class and interface initialization methods (§2.9) are called
        implicitly by the Java Virtual Machine. The value of their
        access_flags item is ignored except for the setting of the ACC_STRICT
        flag.

        All bits of the access_flags item not assigned in Table 4.5 are
        reserved for future use. They should be set to zero in generated class
        files and should be ignored by Java Virtual Machine implementations.
        """
        return (
            (self.ACC_PUBLIC if self.public else 0) |
            (self.ACC_PRIVATE if self.private else 0) |
            (self.ACC_PROTECTED if self.protected else 0) |
            (self.ACC_STATIC if self.static else 0) |
            (self.ACC_FINAL if self.final else 0) |
            (self.ACC_SYNCHRONIZED if self.synchronized else 0) |
            (self.ACC_BRIDGE if self.bridge else 0) |
            (self.ACC_VARARGS if self.varargs else 0) |
            (self.ACC_NATIVE if self.native else 0) |
            (self.ACC_ABSTRACT if self.abstract else 0) |
            (self.ACC_STRICT if self.strict else 0) |
            (self.ACC_SYNTHETIC if self.synthetic else 0)
        )



import codecs
import unittest

from voc.java import mutf_8


codecs.register(mutf_8.search_function)

try:
    # For some reason, Debian feels a need to strip out the test module
    # from packaged versions of Python. It can be manually installed
    # with:
    #   apt-get install libpython3.4-testsuite
    # but if that library hasn't been installed, we can vendor in the
    # required parts.
    from test.test_codecs import ReadTest
except ImportError:
    import io

    class Queue(object):
        """
        queue: write bytes at one end, read bytes from the other end
        """
        def __init__(self, buffer):
            self._buffer = buffer

        def write(self, chars):
            self._buffer += chars

        def read(self, size=-1):
            if size < 0:
                s = self._buffer
                self._buffer = self._buffer[:0]  # make empty
                return s
            else:
                s = self._buffer[:size]
                self._buffer = self._buffer[size:]
                return s

    class MixInCheckStateHandling:
        def check_state_handling_decode(self, encoding, u, s):
            for i in range(len(s)+1):
                d = codecs.getincrementaldecoder(encoding)()
                part1 = d.decode(s[:i])
                state = d.getstate()
                self.assertIsInstance(state[1], int)
                # Check that the condition stated in the documentation for
                # IncrementalDecoder.getstate() holds
                if not state[1]:
                    # reset decoder to the default state without anything buffered
                    d.setstate((state[0][:0], 0))
                    # Feeding the previous input may not produce any output
                    self.assertTrue(not d.decode(state[0]))
                    # The decoder must return to the same state
                    self.assertEqual(state, d.getstate())
                # Create a new decoder and set it to the state
                # we extracted from the old one
                d = codecs.getincrementaldecoder(encoding)()
                d.setstate(state)
                part2 = d.decode(s[i:], True)
                self.assertEqual(u, part1+part2)

        def check_state_handling_encode(self, encoding, u, s):
            for i in range(len(u)+1):
                d = codecs.getincrementalencoder(encoding)()
                part1 = d.encode(u[:i])
                state = d.getstate()
                d = codecs.getincrementalencoder(encoding)()
                d.setstate(state)
                part2 = d.encode(u[i:], True)
                self.assertEqual(s, part1+part2)

    class ReadTest(MixInCheckStateHandling):
        def check_partial(self, input, partialresults):
            # get a StreamReader for the encoding and feed the bytestring version
            # of input to the reader byte by byte. Read everything available from
            # the StreamReader and check that the results equal the appropriate
            # entries from partialresults.
            q = Queue(b"")
            r = codecs.getreader(self.encoding)(q)
            result = ""
            for (c, partialresult) in zip(input.encode(self.encoding), partialresults):
                q.write(bytes([c]))
                result += r.read()
                self.assertEqual(result, partialresult)
            # check that there's nothing left in the buffers
            self.assertEqual(r.read(), "")
            self.assertEqual(r.bytebuffer, b"")

            # do the check again, this time using an incremental decoder
            d = codecs.getincrementaldecoder(self.encoding)()
            result = ""
            for (c, partialresult) in zip(input.encode(self.encoding), partialresults):
                result += d.decode(bytes([c]))
                self.assertEqual(result, partialresult)
            # check that there's nothing left in the buffers
            self.assertEqual(d.decode(b"", True), "")
            self.assertEqual(d.buffer, b"")

            # Check whether the reset method works properly
            d.reset()
            result = ""
            for (c, partialresult) in zip(input.encode(self.encoding), partialresults):
                result += d.decode(bytes([c]))
                self.assertEqual(result, partialresult)
            # check that there's nothing left in the buffers
            self.assertEqual(d.decode(b"", True), "")
            self.assertEqual(d.buffer, b"")

            # check iterdecode()
            encoded = input.encode(self.encoding)
            self.assertEqual(
                input,
                "".join(codecs.iterdecode([bytes([c]) for c in encoded], self.encoding))
            )

        def test_readline(self):
            def getreader(input):
                stream = io.BytesIO(input.encode(self.encoding))
                return codecs.getreader(self.encoding)(stream)

            def readalllines(input, keepends=True, size=None):
                reader = getreader(input)
                lines = []
                while True:
                    line = reader.readline(size=size, keepends=keepends)
                    if not line:
                        break
                    lines.append(line)
                return "|".join(lines)

            s = "foo\nbar\r\nbaz\rspam\u2028eggs"
            sexpected = "foo\n|bar\r\n|baz\r|spam\u2028|eggs"
            sexpectednoends = "foo|bar|baz|spam|eggs"
            self.assertEqual(readalllines(s, True), sexpected)
            self.assertEqual(readalllines(s, False), sexpectednoends)
            self.assertEqual(readalllines(s, True, 10), sexpected)
            self.assertEqual(readalllines(s, False, 10), sexpectednoends)

            lineends = ("\n", "\r\n", "\r", "\u2028")
            # Test long lines (multiple calls to read() in readline())
            vw = []
            vwo = []
            for (i, lineend) in enumerate(lineends):
                vw.append((i*200+200)*"\u3042" + lineend)
                vwo.append((i*200+200)*"\u3042")
            self.assertEqual(readalllines("".join(vw), True), "|".join(vw))
            self.assertEqual(readalllines("".join(vw), False), "|".join(vwo))

            # Test lines where the first read might end with \r, so the
            # reader has to look ahead whether this is a lone \r or a \r\n
            for size in range(80):
                for lineend in lineends:
                    s = 10*(size*"a" + lineend + "xxx\n")
                    reader = getreader(s)
                    for i in range(10):
                        self.assertEqual(
                            reader.readline(keepends=True),
                            size*"a" + lineend,
                        )
                        self.assertEqual(
                            reader.readline(keepends=True),
                            "xxx\n",
                        )
                    reader = getreader(s)
                    for i in range(10):
                        self.assertEqual(
                            reader.readline(keepends=False),
                            size*"a",
                        )
                        self.assertEqual(
                            reader.readline(keepends=False),
                            "xxx",
                        )

        def test_mixed_readline_and_read(self):
            lines = ["Humpty Dumpty sat on a wall,\n",
                     "Humpty Dumpty had a great fall.\r\n",
                     "All the king's horses and all the king's men\r",
                     "Couldn't put Humpty together again."]
            data = ''.join(lines)

            def getreader():
                stream = io.BytesIO(data.encode(self.encoding))
                return codecs.getreader(self.encoding)(stream)

            # Issue #8260: Test readline() followed by read()
            f = getreader()
            self.assertEqual(f.readline(), lines[0])
            self.assertEqual(f.read(), ''.join(lines[1:]))
            self.assertEqual(f.read(), '')

            # Issue #16636: Test readline() followed by readlines()
            f = getreader()
            self.assertEqual(f.readline(), lines[0])
            self.assertEqual(f.readlines(), lines[1:])
            self.assertEqual(f.read(), '')

            # Test read() followed by read()
            f = getreader()
            self.assertEqual(f.read(size=40, chars=5), data[:5])
            self.assertEqual(f.read(), data[5:])
            self.assertEqual(f.read(), '')

            # Issue #12446: Test read() followed by readlines()
            f = getreader()
            self.assertEqual(f.read(size=40, chars=5), data[:5])
            self.assertEqual(f.readlines(), [lines[0][5:]] + lines[1:])
            self.assertEqual(f.read(), '')

        def test_bug1175396(self):
            s = [
                '<%!--===================================================\r\n',
                '    BLOG index page: show recent articles,\r\n',
                '    today\'s articles, or articles of a specific date.\r\n',
                '========================================================--%>\r\n',
                '<%@inputencoding="ISO-8859-1"%>\r\n',
                '<%@pagetemplate=TEMPLATE.y%>\r\n',
                '<%@import=import frog.util, frog%>\r\n',
                '<%@import=import frog.objects%>\r\n',
                '<%@import=from frog.storageerrors import StorageError%>\r\n',
                '<%\r\n',
                '\r\n',
                'import logging\r\n',
                'log=logging.getLogger("Snakelets.logger")\r\n',
                '\r\n',
                '\r\n',
                'user=self.SessionCtx.user\r\n',
                'storageEngine=self.SessionCtx.storageEngine\r\n',
                '\r\n',
                '\r\n',
                'def readArticlesFromDate(date, count=None):\r\n',
                '    entryids=storageEngine.listBlogEntries(date)\r\n',
                '    entryids.reverse() # descending\r\n',
                '    if count:\r\n',
                '        entryids=entryids[:count]\r\n',
                '    try:\r\n',
                '        return [ frog.objects.BlogEntry.load(storageEngine, date, Id) for Id in entryids ]\r\n',
                '    except StorageError,x:\r\n',
                '        log.error("Error loading articles: "+str(x))\r\n',
                '        self.abort("cannot load articles")\r\n',
                '\r\n',
                'showdate=None\r\n',
                '\r\n',
                'arg=self.Request.getArg()\r\n',
                'if arg=="today":\r\n',
                '    #-------------------- TODAY\'S ARTICLES\r\n',
                '    self.write("<h2>Today\'s articles</h2>")\r\n',
                '    showdate = frog.util.isodatestr() \r\n',
                '    entries = readArticlesFromDate(showdate)\r\n',
                'elif arg=="active":\r\n',
                '    #-------------------- ACTIVE ARTICLES redirect\r\n',
                '    self.Yredirect("active.y")\r\n',
                'elif arg=="login":\r\n',
                '    #-------------------- LOGIN PAGE redirect\r\n',
                '    self.Yredirect("login.y")\r\n',
                'elif arg=="date":\r\n',
                '    #-------------------- ARTICLES OF A SPECIFIC DATE\r\n',
                '    showdate = self.Request.getParameter("date")\r\n',
                '    self.write("<h2>Articles written on %s</h2>"% frog.util.mediumdatestr(showdate))\r\n',
                '    entries = readArticlesFromDate(showdate)\r\n',
                'else:\r\n',
                '    #-------------------- RECENT ARTICLES\r\n',
                '    self.write("<h2>Recent articles</h2>")\r\n',
                '    dates=storageEngine.listBlogEntryDates()\r\n',
                '    if dates:\r\n',
                '        entries=[]\r\n',
                '        SHOWAMOUNT=10\r\n',
                '        for showdate in dates:\r\n',
                '            entries.extend( readArticlesFromDate(showdate, SHOWAMOUNT-len(entries)) )\r\n',
                '            if len(entries)>=SHOWAMOUNT:\r\n',
                '                break\r\n',
                '                \r\n',
            ]
            stream = io.BytesIO("".join(s).encode(self.encoding))
            reader = codecs.getreader(self.encoding)(stream)
            for (i, line) in enumerate(reader):
                self.assertEqual(line, s[i])

        def test_readlinequeue(self):
            q = Queue(b"")
            writer = codecs.getwriter(self.encoding)(q)
            reader = codecs.getreader(self.encoding)(q)

            # No lineends
            writer.write("foo\r")
            self.assertEqual(reader.readline(keepends=False), "foo")
            writer.write("\nbar\r")
            self.assertEqual(reader.readline(keepends=False), "")
            self.assertEqual(reader.readline(keepends=False), "bar")
            writer.write("baz")
            self.assertEqual(reader.readline(keepends=False), "baz")
            self.assertEqual(reader.readline(keepends=False), "")

            # Lineends
            writer.write("foo\r")
            self.assertEqual(reader.readline(keepends=True), "foo\r")
            writer.write("\nbar\r")
            self.assertEqual(reader.readline(keepends=True), "\n")
            self.assertEqual(reader.readline(keepends=True), "bar\r")
            writer.write("baz")
            self.assertEqual(reader.readline(keepends=True), "baz")
            self.assertEqual(reader.readline(keepends=True), "")
            writer.write("foo\r\n")
            self.assertEqual(reader.readline(keepends=True), "foo\r\n")

        def test_bug1098990_a(self):
            s1 = "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy\r\n"
            s2 = "offending line: ladfj askldfj klasdj fskla dfzaskdj fasklfj laskd fjasklfzzzzaa%whereisthis!!!\r\n"
            s3 = "next line.\r\n"

            s = (s1+s2+s3).encode(self.encoding)
            stream = io.BytesIO(s)
            reader = codecs.getreader(self.encoding)(stream)
            self.assertEqual(reader.readline(), s1)
            self.assertEqual(reader.readline(), s2)
            self.assertEqual(reader.readline(), s3)
            self.assertEqual(reader.readline(), "")

        def test_bug1098990_b(self):
            s1 = "aaaaaaaaaaaaaaaaaaaaaaaa\r\n"
            s2 = "bbbbbbbbbbbbbbbbbbbbbbbb\r\n"
            s3 = "stillokay:bbbbxx\r\n"
            s4 = "broken!!!!badbad\r\n"
            s5 = "againokay.\r\n"

            s = (s1+s2+s3+s4+s5).encode(self.encoding)
            stream = io.BytesIO(s)
            reader = codecs.getreader(self.encoding)(stream)
            self.assertEqual(reader.readline(), s1)
            self.assertEqual(reader.readline(), s2)
            self.assertEqual(reader.readline(), s3)
            self.assertEqual(reader.readline(), s4)
            self.assertEqual(reader.readline(), s5)
            self.assertEqual(reader.readline(), "")

        ill_formed_sequence_replace = "\ufffd"

        def test_lone_surrogates(self):
            self.assertRaises(UnicodeEncodeError, "\ud800".encode, self.encoding)
            self.assertEqual("[\uDC80]".encode(self.encoding, "backslashreplace"),
                             "[\\udc80]".encode(self.encoding))
            self.assertEqual("[\uDC80]".encode(self.encoding, "namereplace"),
                             "[\\udc80]".encode(self.encoding))
            self.assertEqual("[\uDC80]".encode(self.encoding, "xmlcharrefreplace"),
                             "[&#56448;]".encode(self.encoding))
            self.assertEqual("[\uDC80]".encode(self.encoding, "ignore"),
                             "[]".encode(self.encoding))
            self.assertEqual("[\uDC80]".encode(self.encoding, "replace"),
                             "[?]".encode(self.encoding))

            # sequential surrogate characters
            self.assertEqual("[\uD800\uDC80]".encode(self.encoding, "ignore"),
                             "[]".encode(self.encoding))
            self.assertEqual("[\uD800\uDC80]".encode(self.encoding, "replace"),
                             "[??]".encode(self.encoding))

            bom = "".encode(self.encoding)
            for before, after in [("\U00010fff", "A"), ("[", "]"),
                                  ("A", "\U00010fff")]:
                before_sequence = before.encode(self.encoding)[len(bom):]
                after_sequence = after.encode(self.encoding)[len(bom):]
                test_string = before + "\uDC80" + after
                test_sequence = (bom + before_sequence +
                                 self.ill_formed_sequence + after_sequence)
                self.assertRaises(UnicodeDecodeError, test_sequence.decode,
                                  self.encoding)
                self.assertEqual(test_string.encode(self.encoding,
                                                    "surrogatepass"),
                                 test_sequence)
                self.assertEqual(test_sequence.decode(self.encoding,
                                                      "surrogatepass"),
                                 test_string)
                self.assertEqual(test_sequence.decode(self.encoding, "ignore"),
                                 before + after)
                self.assertEqual(test_sequence.decode(self.encoding, "replace"),
                                 before + self.ill_formed_sequence_replace + after)
                backslashreplace = ''.join('\\x%02x' % b
                                           for b in self.ill_formed_sequence)
                self.assertEqual(test_sequence.decode(self.encoding, "backslashreplace"),
                                 before + backslashreplace + after)


class MUTF8Test(ReadTest, unittest.TestCase):
    encoding = 'mutf-8'

    def test_decoder_state(self):
        u = "\x00\x7f\x80\xff\u0100\u07ff\u0800\uffff\U0010ffff"
        self.check_state_handling_decode(self.encoding,
                                         u, u.encode(self.encoding))

    def test_lone_surrogates(self):
        self.assertRaises(UnicodeDecodeError, b"\xed\xa0\x80".decode, self.encoding)
        self.assertEqual("[\uDC80]".encode(self.encoding, "backslashreplace"),
                         b'[\\udc80]')
        self.assertEqual("[\uDC80]".encode(self.encoding, "xmlcharrefreplace"),
                         b'[&#56448;]')
        self.assertEqual("[\uDC80]".encode(self.encoding, "surrogateescape"),
                         b'[\x80]')
        self.assertEqual("[\uDC80]".encode(self.encoding, "ignore"),
                         b'[]')
        self.assertEqual("[\uDC80]".encode(self.encoding, "replace"),
                         b'[?]')

    def test_surrogatepass_handler(self):
        self.assertEqual("abc\ud800def".encode(self.encoding, "surrogatepass"),
                         b"abc\xed\xa0\x80def")
        self.assertEqual(b"abc\xed\xa0\x80def".decode(self.encoding, "surrogatepass"),
                         "abc\ud800def")
        self.assertTrue(codecs.lookup_error("surrogatepass"))

    def test_invalid(self):
        for invalid in (
            b'\xC0\x81',
            b'\xC0\xFF',
            b'\xC1\x10',
            b'\xC1\x80',
        ):
            with self.assertRaises(UnicodeDecodeError):
                invalid.decode(self.encoding)

    def test_partial(self):
        self.check_partial(
            "\x00\xff\u07ff\u0800\uffff",
            [
                "",
                "\x00",
                "\x00",
                "\x00\xff",
                "\x00\xff",
                "\x00\xff\u07ff",
                "\x00\xff\u07ff",
                "\x00\xff\u07ff",
                "\x00\xff\u07ff\u0800",
                "\x00\xff\u07ff\u0800",
                "\x00\xff\u07ff\u0800",
                "\x00\xff\u07ff\u0800\uffff",
            ]
        )

    def test_null_byte(self):
        self.assertEqual(b'a\xc0\x80b'.decode(self.encoding), 'a\x00b')
        self.assertEqual('a\x00b'.encode(self.encoding), b'a\xc0\x80b')

    def test_surrogates(self):
        self.assertEqual(
            'Hot \U00010400iggity'.encode(self.encoding),
            b'Hot \xed\xa0\x81\xed\xb0\x80iggity',
            'no match'
        )
        self.assertEqual(
            'How \u0205ccentric'.encode(self.encoding),
            b'How \xc8\x85ccentric'
        )

        self.assertEqual(
            b'Hot \xed\xa0\x81\xed\xb0\x80iggity'.decode(self.encoding),
            'Hot \U00010400iggity',
            'no match'
        )
        self.assertEqual(
            b'How \xc8\x85ccentric'.decode(self.encoding),
            'How \u0205ccentric'
        )



from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase


class ComplexTests(TranspileTestCase):
    def test_conjugate(self):
        self.assertCodeExecution("""
            x = complex(1.5, 1)
            print(x.conjugate())
            """)

    def test_real_imag(self):
        self.assertCodeExecution("""
            x = complex(1, 2.0)
            print(x.real)
            print(x.imag)
            """)

    def test_equality_with_numbers_when_zero_imag(self):
        self.assertCodeExecution("""
            x = 2
            y = complex(2, 0)
            print(x == y)
            print(y.__eq__(x))
            print(x != y)
            print(y.__ne__(x))

            x = 2.0
            y = complex(3, 0)
            print(x == y)
            print(y.__eq__(x))
            print(x != y)
            print(y.__ne__(x))

            x = True
            y = complex(1, 0)
            print(x == y)
            print(y.__eq__(x))
            print(x != y)
            print(y.__ne__(x))
            """)


class UnaryComplexOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'complex'


class BinaryComplexOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'complex'

    substitutions = {
        "(-161.18751321137705+195.77962956590406j)": [
            "(-161.18751321137705+195.77962956590403j)"
        ],
        "(2.6460893340172016e-18+0.04321391826377225j)": [
            "(2.6460019439688186e-18+0.04321391826377225j)"
        ],
        "(-9.8368221286278e-14-535.4916555247646j)": [
            "(-9.836497256617357e-14-535.4916555247646j)"
        ]
    }

    is_flakey = [
        'test_power_complex',
        'test_power_float',
    ]


class InplaceComplexOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'complex'

    not_implemented = [
        'test_eq_None',
        'test_eq_NotImplemented',
        'test_eq_slice',
        'test_eq_range',
        'test_eq_bool',
        'test_eq_bytearray',
        'test_eq_bytes',
        'test_eq_class',
        'test_eq_complex',
        'test_eq_dict',
        'test_eq_float',
        'test_eq_frozenset',
        'test_eq_int',
        'test_eq_list',
        'test_eq_none',
        'test_eq_set',
        'test_eq_str',
        'test_eq_tuple',

        'test_ge_None',
        'test_ge_NotImplemented',
        'test_ge_slice',
        'test_ge_range',
        'test_ge_bool',
        'test_ge_bytearray',
        'test_ge_bytes',
        'test_ge_class',
        'test_ge_complex',
        'test_ge_dict',
        'test_ge_float',
        'test_ge_frozenset',
        'test_ge_int',
        'test_ge_list',
        'test_ge_none',
        'test_ge_set',
        'test_ge_str',
        'test_ge_tuple',

        'test_gt_None',
        'test_gt_NotImplemented',
        'test_gt_slice',
        'test_gt_range',
        'test_gt_bool',
        'test_gt_bytearray',
        'test_gt_bytes',
        'test_gt_class',
        'test_gt_complex',
        'test_gt_dict',
        'test_gt_float',
        'test_gt_frozenset',
        'test_gt_int',
        'test_gt_list',
        'test_gt_none',
        'test_gt_set',
        'test_gt_str',
        'test_gt_tuple',

        'test_le_None',
        'test_le_NotImplemented',
        'test_le_slice',
        'test_le_range',
        'test_le_bool',
        'test_le_bytearray',
        'test_le_bytes',
        'test_le_class',
        'test_le_complex',
        'test_le_dict',
        'test_le_float',
        'test_le_frozenset',
        'test_le_int',
        'test_le_list',
        'test_le_none',
        'test_le_set',
        'test_le_str',
        'test_le_tuple',

        'test_lshift_none',

        'test_lt_None',
        'test_lt_NotImplemented',
        'test_lt_slice',
        'test_lt_range',
        'test_lt_bool',
        'test_lt_bytearray',
        'test_lt_bytes',
        'test_lt_class',
        'test_lt_complex',
        'test_lt_dict',
        'test_lt_float',
        'test_lt_frozenset',
        'test_lt_int',
        'test_lt_list',
        'test_lt_none',
        'test_lt_set',
        'test_lt_str',
        'test_lt_tuple',

        'test_multiply_bytearray',
        'test_multiply_bytes',
        'test_multiply_list',
        'test_multiply_none',
        'test_multiply_str',
        'test_multiply_tuple',

        'test_ne_None',
        'test_ne_NotImplemented',
        'test_ne_slice',
        'test_ne_range',
        'test_ne_bool',
        'test_ne_bytearray',
        'test_ne_bytes',
        'test_ne_class',
        'test_ne_complex',
        'test_ne_dict',
        'test_ne_float',
        'test_ne_frozenset',
        'test_ne_int',
        'test_ne_list',
        'test_ne_none',
        'test_ne_set',
        'test_ne_str',
        'test_ne_tuple',

        'test_power_none',

        'test_rshift_none',

        'test_subscr_None',
        'test_subscr_NotImplemented',
        'test_subscr_slice',
        'test_subscr_range',
        'test_subscr_bool',
        'test_subscr_bytearray',
        'test_subscr_bytes',
        'test_subscr_class',
        'test_subscr_complex',
        'test_subscr_dict',
        'test_subscr_float',
        'test_subscr_frozenset',
        'test_subscr_int',
        'test_subscr_list',
        'test_subscr_none',
        'test_subscr_set',
        'test_subscr_str',
        'test_subscr_tuple',
        'test_subscr_obj',

        'test_true_divide_none',
    ]



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class FormatTests(TranspileTestCase):
    pass


class BuiltinFormatFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["format"]

    not_implemented = [
        'test_class',
    ]



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class OrdTests(TranspileTestCase):
    pass


class BuiltinOrdFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["ord"]

    not_implemented = [
        'test_bytes',
        'test_bytearray',
    ]



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class SliceTests(TranspileTestCase):
    def test_slice_repr_stop(self):
        self.assertCodeExecution("""
            print(slice(0))
            print(slice('foo'))
            """)

    def test_slice_repr_start_stop(self):
        self.assertCodeExecution("""
            print(slice(0, 100))
            print(slice('foo', Exception))
            """)

    def test_slice_repr_start_stop_step(self):
        self.assertCodeExecution("""
            print(slice(0, 100, 2))
            print(slice('foo', Exception, object()))
            """)


class BuiltinSliceFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["slice"]

    not_implemented = [
        'test_class',
        'test_complex',
    ]



from unittest import expectedFailure

from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class GlobalsTests(TranspileTestCase):
    def test_simple(self):
        self.assertCodeExecution("""
            print("There are %s globals" % len(globals()))
            x = 1
            y = 'z'
            print("There are %s globals" % len(globals()))

            def method():
                print("In method: there are %s globals" % len(globals()))

                print("globals()['x'] =", globals()['x'])
                print("globals()['y'] =", globals()['y'])
                try:
                    print("globals()['z'] =", globals()['z'])
                except KeyError:
                    print("Variable z not defined")

                globals()[y] = 2

                print("In method: there are %s globals" % len(globals()))

            method()

            print("There are %s globals" % len(globals()))
            print("globals()['x'] =", globals()['x'])
            print("globals()['y'] =", globals()['y'])
            print("globals()['z'] =", globals()['z'])
            print('x', x)
            print('y', y)
            print('z', z)

            print('Done')
        """, run_in_function=False)

    @expectedFailure
    def test_preloaded_globals(self):
        self.assertCodeExecution("""
            glbs = globals
            x = 1
            y = 2

            for i in range(10):
                print(glbs())
        """)

class BuiltinGlobalsFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["globals"]

    not_implemented = [
        'test_bool',
        'test_bytearray',
        'test_bytes',
        'test_class',
        'test_complex',
        'test_dict',
        'test_float',
        'test_frozenset',
        'test_int',
        'test_list',
        'test_None',
        'test_NotImplemented',
        'test_range',
        'test_set',
        'test_slice',
        'test_str',
        'test_tuple',
        'test_obj',
    ]



import ast
import os
import sys

from .python.ast import Visitor
from .python.debug import dump


def transpile(input, prefix='.', outdir=None, namespace='python', verbosity=0):
    transpiler = Transpiler(namespace=namespace, verbosity=verbosity)

    for file_or_dir in input:
        if os.path.isfile(file_or_dir):
            if verbosity:
                print("Compiling %s ..." % file_or_dir)

            with open(file_or_dir, encoding='utf-8') as source:
                ast_module = ast.parse(source.read(), mode='exec')
                transpiler.transpile(file_or_dir, ast_module, prefix)
        elif os.path.isdir(file_or_dir):
            for root, dirs, files in os.walk(file_or_dir, followlinks=True):
                for filename in files:
                    if os.path.splitext(filename)[1] == '.py':
                        source_file = os.path.join(root, filename)
                        if verbosity:
                            print("Compiling %s ..." % source_file)
                        with open(source_file, encoding='utf-8') as source:
                            ast_module = ast.parse(source.read(), mode='exec')
                            transpiler.transpile(source_file, ast_module, prefix)
        else:
            print("Unknown source file: %s" % file_or_dir, file=sys.stderr)

    transpiler.write(outdir)


class Transpiler:
    def __init__(self, namespace="python", verbosity=0):
        self.namespace = namespace
        self.classfiles = []
        self.verbosity = verbosity

    def write(self, outdir):
        # Create directory tree to store classfile
        if outdir:
            basedir = [outdir]
        else:
            basedir = []

        for namespace, class_name, javaclassfile in self.classfiles:
            dirname = os.path.join(*(basedir + namespace.split('.')))
            classfilename = os.path.join(dirname, '%s.class' % class_name)

            try:
                os.makedirs(os.path.dirname(classfilename))
            except FileExistsError:
                pass

            if self.verbosity:
                print("Writing %s ..." % classfilename)

            with open(classfilename, 'wb') as out:
                javaclassfile.write(out)

    def transpile(self, filename, ast_module, prefix):
        "Transpile a Python source file into class files"
        # Determine what portion of the filename is part of the
        # common source directory, and which is namespace.
        common = os.path.commonprefix([
            os.path.abspath(prefix),
            os.path.abspath(filename)
        ])

        self.transpile_code(os.path.abspath(filename)[len(common) + 1:], ast_module)

    def transpile_string(self, filename, code_string):
        "Transpile a string containing Python code into class files"
        ast_module = ast.parse(code_string, mode='exec')
        self.transpile_code(filename, ast_module)

    def transpile_code(self, filename, ast_module):
        "Transpile a code object into class files"
        # Convert the AST into Java opcodes
        if self.verbosity > 1:
            print('=' * 75)
            print(dump(ast_module))
            print('=' * 75)

        module = Visitor(self.namespace, filename, verbosity=self.verbosity).visit(ast_module)

        # Transpile the module code, adding any classfiles generated
        # to the list to be exported.
        self.classfiles.extend(module.transpile())



from unittest import expectedFailure

from ..utils import TranspileTestCase


class ClosureTests(TranspileTestCase):
    def test_deep_nested(self):
        self.assertCodeExecution("""
            def level3():
                def level2():
                    def level1():
                        print("hello")
                    level1()
                level2()
            level3()
            """, run_in_function=False)

    def test_rebind_closure_var_before_closure_construction(self):
        self.assertCodeExecution("""
            def func():
                closure_var = 'before nested is defined'
                def nested():
                    print(closure_var)
                closure_var = 'after nested is defined'
                nested()
            func()
            """)

    def test_generator_closure(self):
        self.assertCodeExecution("""
            def func():
                closure_var = 'hello world'
                def gen():
                    print(closure_var)
                    yield
                next(gen())
            func()

            def gen():
                closure_var = 'hello world'
                def func():
                    print(closure_var)
                func()
                yield
            next(gen())
            """)

    @expectedFailure
    def test_class_closure(self):
        self.assertCodeExecution("""
            def func():
                closure_var = 'hello world'
                class Inner:
                    print(closure_var)
                Inner()
            func()

            def gen():
                closure_var = 'hello world'
                class Inner:
                    print(closure_var)
                Inner()
                yield
            next(gen())
            """)

    @expectedFailure
    def test_method_closure(self):
        self.assertCodeExecution("""
            def func():
                closure_var = 'hello world'
                class Inner:
                    def method(self):
                        print(closure_var)
                Inner().method()
            func()

            def gen():
                closure_var = 'hello world'
                class Inner:
                    def method(self):
                        print(closure_var)
                Inner().method()
                yield
            next(gen())
            """)



from ...java import opcodes as JavaOpcodes, Classref as JavaClassref

from . import java as Java


##########################################################################
# Python types and their operations
##########################################################################

class Callable:
    class invoke:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'org/python/Callable',
                    'invoke',
                    args=['[Lorg/python/Object;', 'Ljava/util/Map;'],
                    returns='Lorg/python/Object;'
                ),
            )


class Iterable:
    class next:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'org/python/Object',
                    '__next__',
                    args=[],
                    returns='Lorg/python/Object;'
                )
            )


class Type:
    class for_class:
        def __init__(self, name):
            self.name = name

        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.LDC_W(JavaClassref(self.name)),
                JavaOpcodes.INVOKESTATIC(
                    'org/python/types/Type',
                    'pythonType',
                    args=['Ljava/lang/Class;'],
                    returns='Lorg/python/types/Type;'
                )
            )

    class for_name:
        def __init__(self, name):
            self.name = name

        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.LDC_W(self.name),
                JavaOpcodes.INVOKESTATIC(
                    'org/python/types/Type',
                    'pythonType',
                    args=['Ljava/lang/String;'],
                    returns='Lorg/python/types/Type;'
                )
            )

    class to_python:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKESTATIC(
                    'org/python/types/Type',
                    'toPython',
                    args=['Ljava/lang/Object;'],
                    returns='Lorg/python/Object;'
                )
            )


class Object:
    class get_attribute:
        def __init__(self, attr, use_null=False):
            self.attr = attr
            self.use_null = use_null

        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.LDC_W(self.attr),
            )
            if self.use_null:
                context.add_opcodes(
                    JavaOpcodes.INVOKEINTERFACE(
                        'org/python/Object',
                        '__getattribute_null',
                        args=['Ljava/lang/String;'],
                        returns='Lorg/python/Object;'
                    ),
                )
            else:
                context.add_opcodes(
                    JavaOpcodes.INVOKEINTERFACE(
                        'org/python/Object',
                        '__getattribute__',
                        args=['Ljava/lang/String;'],
                        returns='Lorg/python/Object;'
                    ),
                )

    class set_attr:
        def __init__(self, attr):
            self.attr = attr

        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.LDC_W(self.attr),
                JavaOpcodes.SWAP(),
                JavaOpcodes.INVOKEINTERFACE(
                    'org/python/Object',
                    '__setattr__',
                    args=['Ljava/lang/String;', 'Lorg/python/Object;'],
                    returns='V'
                ),
            )

    class del_attr:
        def __init__(self, attr=None):
            self.attr = attr

        def process(self, context):
            if self.attr:
                context.add_opcodes(
                    JavaOpcodes.LDC_W(self.attr),
                )
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'org/python/Object',
                    '__delattr__',
                    args=['Ljava/lang/String;'],
                    returns='V'
                ),
            )

    class get_item:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'org/python/Object',
                    '__getitem__',
                    args=['Lorg/python/Object;'],
                    returns='Lorg/python/Object;'
                ),
            )

    class set_item:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'org/python/Object',
                    '__setitem__',
                    args=['Lorg/python/Object;', 'Lorg/python/Object;'],
                    returns='V'
                ),
            )

    class del_item:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'org/python/Object',
                    '__delitem__',
                    args=['Lorg/python/Object;'],
                    returns='V'
                ),
            )

    class iter:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'org/python/Object',
                    '__iter__',
                    args=[],
                    returns='Lorg/python/Object;'
                )
            )

    class as_boolean:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'org/python/Object',
                    'toBoolean',
                    args=[],
                    returns='Z'
                ),
            )


class NONE:
    def process(self, context):
        context.add_opcodes(
            JavaOpcodes.GETSTATIC('org/python/types/NoneType', 'NONE', 'Lorg/python/Object;')
        )


class Dict:
    def process(self, context):
        context.add_opcodes(
            Java.New('org/python/types/Dict'),
            Java.Init('org/python/types/Dict')
        )

    class set_item:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEVIRTUAL(
                    'org/python/types/Dict',
                    '__setitem__',
                    args=['Lorg/python/Object;', 'Lorg/python/Object;'],
                    returns='V'
                )
            )


class Set:
    def process(self, context):
        context.add_opcodes(
            Java.New('org/python/types/Set'),
            Java.Init('org/python/types/Set')
        )

    class add:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEVIRTUAL(
                    'org/python/types/Set',
                    'add',
                    args=['Lorg/python/Object;'],
                    returns='Lorg/python/Object;'
                ),
                JavaOpcodes.POP()
            )


class List:
    def process(self, context):
        context.add_opcodes(
            Java.New('org/python/types/List'),
            Java.Init('org/python/types/List')
        )

    class append:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEVIRTUAL(
                    'org/python/types/List',
                    'append',
                    args=['Lorg/python/Object;'],
                    returns='Lorg/python/Object;'
                ),
                JavaOpcodes.POP()
            )


class Str:
    def __init__(self, value=None):
        self.value = value

    def process(self, context):
        if self.value:
            context.add_opcodes(
                Java.New('org/python/types/Str'),
                JavaOpcodes.LDC_W(self.value),
                Java.Init('org/python/types/Str', 'Ljava/lang/String;')
            )
        else:
            context.add_opcodes(
                Java.New('org/python/types/Str'),
                Java.Init('org/python/types/Str')
            )

class Array:
    class get_item:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'org/python/Object',
                    '__getitem__',
                    args=['I'],
                    returns='Lorg/python/Object;'
                ),
            )

    class set_item:
        def process(self, context):
            context.add_opcodes(
                JavaOpcodes.INVOKEINTERFACE(
                    'org/python/Object',
                    '__setitem__',
                    args=['I', 'Lorg/python/Object;'],
                    returns='V'
                ),
            )



from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase


class NotImplementedTests(TranspileTestCase):
    def test_truth(self):
        self.assertCodeExecution("""
            x = NotImplemented
            print(x == True)
            """)


class UnaryNotImplementedOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'NotImplemented'


class BinaryNotImplementedOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'NotImplemented'


class InplaceNotImplementedOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'NotImplemented'



from ...java import opcodes as JavaOpcodes


##########################################################################
# Local variables are stored in a dictionary, keyed by name,
# and with the value of the local variable register they are stored in.
#
# If an attempt is used to use a name that has been deleted, an exception
# is raised.
#
# Once a name has been deleted, the index will be recycled for re-use
# as a different name.
##########################################################################

class ALOAD_index:
    """Generate the opcode to load an object variable at the given index onto the stack.

    This uses the optimized register operations for the first 4 local variables.
    """
    def __init__(self, index):
        self.index = index

    def process(self, context):
        # if self.index is None:
        #     print("LOAD AVAR NAME", context, self.name)
        #     print("locals: ", context.local_vars, context.deleted_vars)
        if self.index is None:
            raise NameError(self.name)
        elif self.index == 0:
            context.add_opcodes(JavaOpcodes.ALOAD_0())
        elif self.index == 1:
            context.add_opcodes(JavaOpcodes.ALOAD_1())
        elif self.index == 2:
            context.add_opcodes(JavaOpcodes.ALOAD_2())
        elif self.index == 3:
            context.add_opcodes(JavaOpcodes.ALOAD_3())
        else:
            context.add_opcodes(JavaOpcodes.ALOAD(self.index))

        # This opcode isn't for the final output.
        return False


class ALOAD_name(ALOAD_index):
    """Generate the opcode to load an object variable with the given name onto the stack.

    This looks up the local variable dictionary to find which
    register is being used for that variable, using the optimized
    register operations for the first 4 local variables.
    """
    def __init__(self, name):
        self.name = name
        super().__init__(None)

    def process(self, context):
        # print("LOAD AVAR NAME", context, self.name)
        # print("locals: ", context.local_vars, context.deleted_vars)
        try:
            self.index = context.local_vars[self.name]
        except KeyError:
            pass

        return super().process(context)


class ASTORE_index:
    """Generate the opcode to store an object variable at the given index.

    This uses the optimized register operations for the first 4 local variables.
    """
    def __init__(self, index):
        self.index = index

    def process(self, context):
        if self.index is None:
            try:
                self.index = context.deleted_vars.pop()
                # print ("REUSE index", self.index)
            except KeyError:
                self.index = len(context.active_local_vars)
                # print ("GET NEW index", self.index)
            context.local_vars[self.name] = self.index

        # if self.index is None:
        #     print("STORE AVAR NAME", context, self.index, self.name)
        #     print("locals: ", context.local_vars, context.deleted_vars)

        if self.index == 0:
            context.add_opcodes(JavaOpcodes.ASTORE_0())
        elif self.index == 1:
            context.add_opcodes(JavaOpcodes.ASTORE_1())
        elif self.index == 2:
            context.add_opcodes(JavaOpcodes.ASTORE_2())
        elif self.index == 3:
            context.add_opcodes(JavaOpcodes.ASTORE_3())
        else:
            context.add_opcodes(JavaOpcodes.ASTORE(self.index))

        # This opcode isn't for the final output.
        return False


class ASTORE_name(ASTORE_index):
    """Generate the opcode to store an object variable with the given name.

    This looks up the local variable dictionary to find which
    register is being used for that variable, using the optimized
    register operations for the first 4 local variables.
    """
    def __init__(self, name):
        self.name = name
        super().__init__(None)

    def process(self, context):
        try:
            self.index = context.local_vars[self.name]
        except KeyError:
            self.index = None

        return super().process(context)


class ILOAD_name:
    """Generate the opcode to load an integer variable with the given name onto the stack.

    This looks up the local variable dictionary to find which
    register is being used for that variable, using the optimized
    register operations for the first 4 local variables.
    """
    def __init__(self, name):
        self.name = name

    def process(self, context):
        # print("LOAD IVAR NAME", context, self.name)
        # print("locals: ", context.local_vars)
        try:
            index = context.local_vars[self.name]
        except KeyError:
            index = None

        if index is None:
            raise NameError(self.name)
        elif index == 0:
            context.add_opcodes(JavaOpcodes.ILOAD_0())
        elif index == 1:
            context.add_opcodes(JavaOpcodes.ILOAD_1())
        elif index == 2:
            context.add_opcodes(JavaOpcodes.ILOAD_2())
        elif index == 3:
            context.add_opcodes(JavaOpcodes.ILOAD_3())
        else:
            context.add_opcodes(JavaOpcodes.ILOAD(index))

        # This opcode isn't for the final output.
        return False


class ISTORE_name:
    """Generate the opcode to store a variable with the given name.

    This looks up the local variable dictionary to find which
    register is being used for that variable, using the optimized
    register operations for the first 4 local variables.
    """
    def __init__(self, name):
        self.name = name

    def process(self, context):
        try:
            index = context.local_vars[self.name]
        except KeyError:
            index = None

        if index is None:
            try:
                index = context.deleted_vars.pop()
                # print ("REUSE index", index)
            except KeyError:
                index = len(context.active_local_vars)
                # print ("GET NEW index", index)
            context.local_vars[self.name] = index

        # print("STORE IVAR NAME", context, index, self.name)
        # print("locals: ", context.local_vars)

        if index == 0:
            context.add_opcodes(JavaOpcodes.ISTORE_0())
        elif index == 1:
            context.add_opcodes(JavaOpcodes.ISTORE_1())
        elif index == 2:
            context.add_opcodes(JavaOpcodes.ISTORE_2())
        elif index == 3:
            context.add_opcodes(JavaOpcodes.ISTORE_3())
        else:
            context.add_opcodes(JavaOpcodes.ISTORE(index))

        # This opcode isn't for the final output.
        return False


class IINC_name:
    """Generate the opcode to increment an integer variable with the given name
    by the provided value.

    This looks up the local variable dictionary to find which
    register is being used for that variable, using the optimized
    register operations for the first 4 local variables.
    """
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def process(self, context):
        # print("LOAD IVAR NAME", context, self.name)
        # print("locals: ", context.local_vars)
        try:
            index = context.local_vars[self.name]
        except KeyError:
            index = None

        if index is None:
            raise NameError(self.name)

        context.add_opcodes(
            JavaOpcodes.IINC(index, self.value)
        )

        # This opcode isn't for the final output.
        return False


class LLOAD_name:
    """Generate the opcode to load a long variable with the given name onto the stack.

    This looks up the local variable dictionary to find which
    register is being used for that variable, using the optimized
    register operations for the first 4 local variables.
    """
    def __init__(self, name):
        self.name = name

    def process(self, context):
        # print("LOAD LVAR NAME", context, self.name, index)
        # print("locals: ", context.local_vars)
        try:
            index = context.local_vars[self.name]
        except KeyError:
            index = None

        if index is None:
            raise NameError(self.name)
        elif index == 0:
            context.add_opcodes(JavaOpcodes.LLOAD_0())
        elif index == 1:
            context.add_opcodes(JavaOpcodes.LLOAD_1())
        elif index == 2:
            context.add_opcodes(JavaOpcodes.LLOAD_2())
        elif index == 3:
            context.add_opcodes(JavaOpcodes.LLOAD_3())
        else:
            context.add_opcodes(JavaOpcodes.LLOAD(index))

        # This opcode isn't for the final output.
        return False


class FLOAD_name:
    """Generate the opcode to load a float variable with the given name onto the stack.

    This looks up the local variable dictionary to find which
    register is being used for that variable, using the optimized
    register operations for the first 4 local variables.
    """
    def __init__(self, name):
        self.name = name

    def process(self, context):
        # print("LOAD FVAR NAME", context, self.name, index)
        # print("locals: ", context.local_vars)
        try:
            index = context.local_vars[self.name]
        except KeyError:
            index = None

        if index is None:
            raise NameError(self.name)
        elif index == 0:
            context.add_opcodes(JavaOpcodes.FLOAD_0())
        elif index == 1:
            context.add_opcodes(JavaOpcodes.FLOAD_1())
        elif index == 2:
            context.add_opcodes(JavaOpcodes.FLOAD_2())
        elif index == 3:
            context.add_opcodes(JavaOpcodes.FLOAD_3())
        else:
            context.add_opcodes(JavaOpcodes.FLOAD(index))

        # This opcode isn't for the final output.
        return False


class DLOAD_name:
    """Generate the opcode to load a double variable with the given name onto the stack.

    This looks up the local variable dictionary to find which
    register is being used for that variable, using the optimized
    register operations for the first 4 local variables.
    """
    def __init__(self, name):
        self.name = name

    def process(self, context):
        # print("LOAD LVAR NAME", context, self.name, index)
        # print("locals: ", context.local_vars)
        try:
            index = context.local_vars[self.name]
        except KeyError:
            index = None

        if index is None:
            raise NameError(self.name)
        elif index == 0:
            context.add_opcodes(JavaOpcodes.DLOAD_0())
        elif index == 1:
            context.add_opcodes(JavaOpcodes.DLOAD_1())
        elif index == 2:
            context.add_opcodes(JavaOpcodes.DLOAD_2())
        elif index == 3:
            context.add_opcodes(JavaOpcodes.DLOAD_3())
        else:
            context.add_opcodes(JavaOpcodes.DLOAD(index))

        # This opcode isn't for the final output.
        return False


class free_name:
    """Remove a name from the local variable pool

    By default the variable must exist. However, if you pass
    in must_exist, the non-existence of the variable will not
    be treated as an error. This is to allow for variables that
    are created as part of looping constructs, and may not be
    created in the case of an empty loop.
    """
    def __init__(self, name, must_exist=True):
        self.name = name
        self.must_exist = must_exist

    def process(self, context):
        try:
            index = context.local_vars[self.name]
        except KeyError:
            index = None

        if index is None and self.must_exist:
            raise NameError(self.name)

        context.deleted_vars.add(index)
        context.local_vars[self.name] = None

        # print("FREE", context, self.name, index)
        # print("locals: ", context.local_vars, context.deleted_vars)

        # This opcode isn't for the final output.
        return False


##########################################################################
# Handle constant values.
#
# There are multiple opcodes to load and retrieve constants. Use the
# best option available at any given time.
##########################################################################

def ICONST_val(value):
    """Write an integer constant onto the stack.

    There are a couple of opcodes that can be used to optimize the
    loading of small integers; use them if possible.
    """
    if isinstance(value, bool):
        if value:
            return JavaOpcodes.ICONST_1()
        else:
            return JavaOpcodes.ICONST_0()
    elif isinstance(value, int):
        if value == 0:
            return JavaOpcodes.ICONST_0()
        elif value == 1:
            return JavaOpcodes.ICONST_1()
        elif value == 2:
            return JavaOpcodes.ICONST_2()
        elif value == 3:
            return JavaOpcodes.ICONST_3()
        elif value == 4:
            return JavaOpcodes.ICONST_4()
        elif value == 5:
            return JavaOpcodes.ICONST_5()
        elif value == -1:
            return JavaOpcodes.ICONST_M1()
        elif -32768 <= value <= 32767:
            return JavaOpcodes.SIPUSH(value)
        elif -2147483648 <= value <= 2147483647:
            return JavaOpcodes.LDC(value)
        else:
            raise RuntimeError("%s is out of integer range" % value)
    else:
        raise RuntimeError("%s is not an integer constant" % value)


def LCONST_val(value):
    """Write an long integer constant onto the stack.

    There are a couple of opcodes that can be used to optimize the
    loading of small longs; use them if possible.
    """
    if isinstance(value, int):
        if value == 0:
            return JavaOpcodes.LCONST_0()
        elif value == 1:
            return JavaOpcodes.LCONST_1()
        else:
            return JavaOpcodes.LDC2_W(value)
    else:
        raise RuntimeError("%s is not a long integer constant" % value)


def FCONST_val(value):
    """Write a float constant onto the stack.

    There are a couple of opcodes that can be used to optimize the
    loading of some floats; use them if possible.
    """
    if isinstance(value, float):
        if value == 0.0:
            return JavaOpcodes.FCONST_0()
        elif value == 1.0:
            return JavaOpcodes.FCONST_1()
        elif value == 2.0:
            return JavaOpcodes.FCONST_2()
        else:
            return JavaOpcodes.LDC_W(value)
    else:
        raise RuntimeError("%s is not a float constant" % value)


def DCONST_val(value):
    """Write an double constant onto the stack.

    There are a couple of opcodes that can be used to optimize the
    loading of some doubles; use them if possible.
    """
    if isinstance(value, float):
        if value == 0.0:
            return JavaOpcodes.DCONST_0()
        elif value == 1.0:
            return JavaOpcodes.DCONST_1()
        else:
            return JavaOpcodes.LDC2_W(value)
    else:
        raise RuntimeError("%s is not a double constant" % value)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class StrTests(TranspileTestCase):
    pass


class BuiltinStrFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["str"]

    not_implemented = [
        'test_class',
    ]



from .signatures import method_descriptor


##########################################################################
# Some utility methods to help compare and hash constants
##########################################################################


def multieq(obj1, obj2, *attrs):
    "Compare two objects using the provided attributes"
    return (
        type(obj1) == type(obj2)
        and all(getattr(obj1, attr) == getattr(obj2, attr) for attr in attrs)
    )


def multihash(obj, *attrs):
    """Generate a hash of the object using the provided attributes.

    31 is a "magic number"; it's the same magic number used by Java
    when hashing multiple objects.
    """
    result = 1
    for attr in attrs:
        result = 31 * result + hash(getattr(obj, attr))
    return result


##########################################################################
# Some utility methods to help reconstruct a constant pool.
# These allow us to read the entire constant pool, and then resolve it.
##########################################################################

def resolve(entry, pool):
    klass, args = entry

    resolved_args = []
    for arg in args:
        if hasattr(arg, '__call__'):
            resolved_args.append(arg(pool))
        else:
            resolved_args.append(arg)
    return klass(*resolved_args)


def read_ref_attr(index, attr):
    def resolver(pool):
        entry = pool[index - 1]
        if isinstance(entry, tuple):
            value = resolve(entry, pool)
            pool[index - 1] = value
        else:
            value = entry
        return getattr(value, attr).bytes.decode('mutf-8')
    return resolver


def read_utf8(index):
    def resolver(pool):
        entry = pool[index - 1]
        if isinstance(entry, tuple):
            value = resolve(entry, pool)
            pool[index - 1] = value
        else:
            value = entry
        return value.bytes.decode('mutf-8')
    return resolver

# From: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html

##########################################################################
# 4.4. The Constant Pool
##########################################################################

# Java Virtual Machine instructions do not rely on the run-time layout of
# classes, interfaces, class instances, or arrays. Instead, instructions refer
# to symbolic information in the constant_pool table.


class ConstantPool:
    def __init__(self):
        self._constants = {}
        # The constant_pool is a table of structures (§4.4) representing various string
        # constants, class and interface names, field names, and other constants that
        # are referred to within the ClassFile structure and its substructures. The
        # format of each constant_pool table entry is indicated by its first "tag" byte.

        # The constant_pool table is indexed from 1 to constant_pool_count-1.
        self._constant_pool = []

    @property
    def count(self):
        """The value of the constant_pool_count item is equal to the number of entries in
        the constant_pool table plus one. A constant_pool index is considered valid if
        it is greater than zero and less than constant_pool_count, with the exception
        for constants of type long and double noted in §4.4.5.
        """
        return len(self._constant_pool) + 1

    def add(self, obj, allow_duplicates=False):
        if allow_duplicates or obj not in self._constants:
            self._constant_pool.append(obj)
            self._constants[obj] = len(self)
            if obj.__class__ in (Double, Long):
                self._constant_pool.append(None)

    def __len__(self):
        return len(self._constant_pool)

    def __getitem__(self, i):
        return self._constant_pool[i - 1]

    def index(self, obj):
        return self._constants[obj]

    def read(self, reader, dump=None):
        count = reader.read_u2()
        if dump is not None:
            reader.debug("    " * dump, 'Constant pool: (%s constants)' % (count - 1))

        raw_pool = []
        i = 1
        while i < count:
            constant = Constant.read(reader)
            raw_pool.append(constant)
            if constant[0] in (Double, Long):
                raw_pool.append(None)
                i += 2
            else:
                i += 1

        i = 0
        while i < count - 1:
            entry = raw_pool[i]
            if isinstance(entry, tuple):
                const = resolve(entry, raw_pool)
            else:
                const = entry
            self.add(const, allow_duplicates=True)
            if dump is not None:
                reader.debug("    " * (dump + 1), '%s: %s' % ((i + 1), repr(const)))
            if const.__class__ in (Double, Long):
                i += 2
            else:
                i += 1

    def write(self, writer):
        writer.write_u2(self.count)
        for i, constant in enumerate(self._constant_pool):
            if constant:
                constant.write(writer)


# All constant_pool table entries have the following general format:

class Constant:
    # u1 tag;
    # u1 info[];

    # Each item in the constant_pool table must begin with a 1-byte tag indicating
    # the kind of cp_info entry. The contents of the info array vary with the value
    # of tag. The valid tags and their values are listed in Table 4.3. Each tag byte
    # must be followed by two or more bytes giving information about the specific
    # constant. The format of the additional information varies with the tag value.

    # Table 4.3. Constant pool tags

    CONSTANT_Class = 7
    CONSTANT_Fieldref = 9
    CONSTANT_Methodref = 10
    CONSTANT_InterfaceMethodref = 11
    CONSTANT_String = 8
    CONSTANT_Integer = 3
    CONSTANT_Float = 4
    CONSTANT_Long = 5
    CONSTANT_Double = 6
    CONSTANT_NameAndType = 12
    CONSTANT_Utf8 = 1
    CONSTANT_MethodHandle = 15
    CONSTANT_MethodType = 16
    CONSTANT_InvokeDynamic = 18

    def __init__(self, tag):
        self.tag = tag

    @staticmethod
    def read(reader):
        tag = reader.read_u1()
        klass = {
            0: None,
            Constant.CONSTANT_Class: Classref,
            Constant.CONSTANT_Fieldref: Fieldref,
            Constant.CONSTANT_Methodref: Methodref,
            Constant.CONSTANT_InterfaceMethodref: InterfaceMethodref,
            Constant.CONSTANT_String: String,
            Constant.CONSTANT_Integer: Integer,
            Constant.CONSTANT_Float: Float,
            Constant.CONSTANT_Long: Long,
            Constant.CONSTANT_Double: Double,
            Constant.CONSTANT_NameAndType: NameAndType,
            Constant.CONSTANT_Utf8: Utf8,
            # Constant.CONSTANT_MethodHandle: MethodHandle
            # Constant.CONSTANT_MethodType: MethodType
            # Constant.CONSTANT_InvokeDynamic: InvokeDynamic
        }[tag]
        if klass:
            return klass.read_info(reader)

    def write(self, writer):
        writer.write_u1(self.tag)
        self.write_info(writer)

    def resolve(self, constant_pool):
        constant_pool.add(self)
        self.resolve_info(constant_pool)

    def resolve_info(self, constant_pool):
        pass

# ------------------------------------------------------------------------
# 4.4.1. The CONSTANT_Class_info Structure
# ------------------------------------------------------------------------

# The CONSTANT_Class_info structure is used to represent a class or an interface:


class Classref(Constant):
    # u1 tag;
    # u2 name_index;

    def __init__(self, name):
        # The tag item has the value CONSTANT_Class (7).
        super().__init__(Constant.CONSTANT_Class)

        # The value of the name item must be a valid index into the
        # constant_pool table. The constant_pool entry at that index must be a
        # CONSTANT_Utf8_info (§4.4.7) structure representing a valid binary
        # class or interface name encoded in internal form (§4.2.1).

        # Because arrays are objects, the opcodes anewarray and multianewarray
        # can reference array "classes" via CONSTANT_Class_info structures in
        # the constant_pool table. For such array classes, the name of the class
        # is the descriptor of the array type.

        # For example, the class name representing a two-dimensional int array
        # type:
        #
        #     int[][]
        #
        # is:
        #
        #     [[I
        #
        # The class name representing the type array of class Thread:
        #
        #     Thread[]
        # is
        #
        #     [Ljava/lang/Thread;
        #
        # An array type descriptor is valid only if it represents 255 or fewer
        # dimensions.
        self.class_name = name
        self.name = Utf8(name)

    def __repr__(self):
        return '<Class %s>' % self.name

    def __eq__(self, other):
        return isinstance(other, Classref) and other.tag == self.tag and other.name == self.name

    def __hash__(self):
        return multihash(self, 'tag', 'name')

    @staticmethod
    def read_info(reader):
        name = reader.read_u2()
        return Classref, (read_utf8(name),)

    def write_info(self, writer):
        writer.write_u2(writer.constant_pool.index(self.name))

    def resolve_info(self, constant_pool):
        self.name.resolve(constant_pool)


# ------------------------------------------------------------------------
# 4.4.2. The CONSTANT_Fieldref_info, CONSTANT_Methodref_info, and CONSTANT_InterfaceMethodref_info Structures
# ------------------------------------------------------------------------

# Fields, methods, and interface methods are represented by similar structures:

class Fieldref(Constant):
    # u1 tag;
    # u2 class_index;
    # u2 name_and_type_index;

    def __init__(self, class_name, name, descriptor):

        # The tag item of a CONSTANT_Fieldref_info structure has the value
        # CONSTANT_Fieldref (9).
        super().__init__(Constant.CONSTANT_Fieldref)

        # The value of the class item must be a valid index into the
        # constant_pool table. The constant_pool entry at that index must be a
        # CONSTANT_Class_info (§4.4.1) structure representing a class or
        # interface type that has the field or method as a member.

        # The class_index item of a CONSTANT_Fieldref_info structure may be
        # either a class type or an interface type.
        self.class_name = class_name
        self.klass = Classref(class_name)

        # The value of the name_and_type item must be a valid index into
        # the constant_pool table. The constant_pool entry at that index must be
        # a CONSTANT_NameAndType_info (§4.4.6) structure. This constant_pool
        # entry indicates the name and descriptor of the field or method.

        # The indicated descriptor must be a field descriptor (§4.3.2).
        self.name = name
        self.descriptor = descriptor
        self.name_and_type = NameAndType(name, descriptor)

    def __repr__(self):
        return '<Fieldref %s.%s (%s)>' % (self.class_name, self.name, self.descriptor)

    def __eq__(self, other):
        return (
            isinstance(other, Fieldref)
            and other.tag == self.tag
            and other.klass == self.klass
            and other.name_and_type == self.name_and_type
        )

    def __hash__(self):
        return multihash(self, 'tag', 'klass', 'name_and_type')

    def __info_eq__(self, other):
        return

    def __info_hash__(self):
        return 31 * hash(self.klass) + 31

    @staticmethod
    def read_info(reader):
        klass = reader.read_u2()
        name_and_type = reader.read_u2()
        return Fieldref, (
            read_ref_attr(klass, 'name'),
            read_ref_attr(name_and_type, 'name'),
            read_ref_attr(name_and_type, 'descriptor'),
        )

    def write_info(self, writer):
        writer.write_u2(writer.constant_pool.index(self.klass))
        writer.write_u2(writer.constant_pool.index(self.name_and_type))

    def resolve_info(self, constant_pool):
        self.klass.resolve(constant_pool)
        self.name_and_type.resolve(constant_pool)


class Methodref(Constant):
    # u1 tag;
    # u2 class_index;
    # u2 name_and_type_index;

    def __init__(self, class_name, name, descriptor):

        # The tag item of a CONSTANT_Methodref_info structure has the value
        # CONSTANT_Methodref (10).
        super().__init__(Constant.CONSTANT_Methodref)

        # The value of the class item must be a valid index into the
        # constant_pool table. The constant_pool entry at that index must be a
        # CONSTANT_Class_info (§4.4.1) structure representing a class or
        # interface type that has the field or method as a member.

        # The class item of a CONSTANT_Methodref_info structure must be a
        # class type, not an interface type.
        self.class_name = class_name
        self.klass = Classref(class_name)

        # The value of the name_and_type item must be a valid index into
        # the constant_pool table. The constant_pool entry at that index must be
        # a CONSTANT_NameAndType_info (§4.4.6) structure. This constant_pool
        # entry indicates the name and descriptor of the field or method.

        # The indicated descriptor must be a method descriptor (§4.3.3).

        # If the name of the method of a CONSTANT_Methodref_info structure
        # begins with a '<' ('\u003c'), then the name must be the special name
        # <init>, representing an instance initialization method (§2.9). The
        # return type of such a method must be void.
        self.name_and_type = NameAndType(name, descriptor)

        # For convenience, store a parsed version of the descriptor.
        self.name = name
        self.descriptor = method_descriptor(descriptor)

    def __repr__(self):
        return '<Methodref %s.%s %s>' % (self.class_name, self.name, self.name_and_type.descriptor)

    def __eq__(self, other):
        return (
            isinstance(other, Methodref)
            and other.tag == self.tag
            and other.klass == self.klass
            and other.name_and_type == self.name_and_type
        )

    def __hash__(self):
        return multihash(self, 'tag', 'klass', 'name_and_type')

    @staticmethod
    def read_info(reader):
        klass = reader.read_u2()
        name_and_type = reader.read_u2()
        return Methodref, (
            read_ref_attr(klass, 'name'),
            read_ref_attr(name_and_type, 'name'),
            read_ref_attr(name_and_type, 'descriptor'),
        )

    def write_info(self, writer):
        writer.write_u2(writer.constant_pool.index(self.klass))
        writer.write_u2(writer.constant_pool.index(self.name_and_type))

    def resolve_info(self, constant_pool):
        self.klass.resolve(constant_pool)
        self.name_and_type.resolve(constant_pool)


class InterfaceMethodref(Constant):
    # u1 tag;
    # u2 class_index;
    # u2 name_and_type_index;

    def __init__(self, class_name, name, descriptor):

        # The tag item of a CONSTANT_InterfaceMethodref_info structure has the value
        # CONSTANT_InterfaceMethodref (11).
        super().__init__(Constant.CONSTANT_InterfaceMethodref)

        # The value of the class item must be a valid index into the
        # constant_pool table. The constant_pool entry at that index must be a
        # CONSTANT_Class_info (§4.4.1) structure representing a class or
        # interface type that has the field or method as a member.

        # The class item of a CONSTANT_InterfaceMethodref_info structure
        # must be an interface type.
        self.class_name = class_name
        self.klass = Classref(class_name)

        # The value of the name_and_type item must be a valid index into
        # the constant_pool table. The constant_pool entry at that index must be
        # a CONSTANT_NameAndType_info (§4.4.6) structure. This constant_pool
        # entry indicates the name and descriptor of the field or method.

        # The indicated descriptor must be a method descriptor (§4.3.3).
        self.name_and_type = NameAndType(name, descriptor)

        # For convenience, store a parsed version of the descriptor.
        self.name = name
        self.descriptor = method_descriptor(descriptor)

    def __repr__(self):
        return '<InterfaceMethodref %s.%s %s>' % (self.class_name, self.name, self.name_and_type.descriptor)

    def __eq__(self, other):
        return (
            isinstance(other, InterfaceMethodref)
            and other.tag == self.tag
            and other.klass == self.klass
            and other.name_and_type == self.name_and_type
        )

    def __hash__(self):
        return multihash(self, 'tag', 'klass', 'name_and_type')

    @staticmethod
    def read_info(reader):
        klass = reader.read_u2()
        name_and_type = reader.read_u2()
        return InterfaceMethodref, (
            read_ref_attr(klass, 'name'),
            read_ref_attr(name_and_type, 'name'),
            read_ref_attr(name_and_type, 'descriptor'),
        )

    def write_info(self, writer):
        writer.write_u2(writer.constant_pool.index(self.klass))
        writer.write_u2(writer.constant_pool.index(self.name_and_type))

    def resolve_info(self, constant_pool):
        self.klass.resolve(constant_pool)
        self.name_and_type.resolve(constant_pool)


# ------------------------------------------------------------------------
# 4.4.3. The CONSTANT_String_info Structure
# ------------------------------------------------------------------------

# The CONSTANT_String_info structure is used to represent constant objects of
# the type String:

class String(Constant):
    # u1 tag;
    # u2 string_index;

    def __init__(self, value):
        # The tag item of the CONSTANT_String_info structure has the value
        # CONSTANT_String (8).
        super().__init__(Constant.CONSTANT_String)

        # The value of the string item must be a valid index into the
        # constant_pool table. The constant_pool entry at that index must be a
        # CONSTANT_Utf8_info (§4.4.7) structure representing the sequence of
        # Unicode code points to which the String object is to be initialized.
        self.value = Utf8(value)

    def __repr__(self):
        return "<String '%s'>" % self.value

    def __eq__(self, other):
        return isinstance(other, String) and other.tag == self.tag and other.value == self.value

    def __hash__(self):
        return multihash(self, 'tag', 'value')

    @staticmethod
    def read_info(reader):
        value = reader.read_u2()
        return String, (
            read_utf8(value),
        )

    def write_info(self, writer):
        writer.write_u2(writer.constant_pool.index(self.value))

    def resolve_info(self, constant_pool):
        self.value.resolve(constant_pool)


# ------------------------------------------------------------------------
# 4.4.4. The CONSTANT_Integer_info and CONSTANT_Float_info Structures
# ------------------------------------------------------------------------

# The CONSTANT_Integer_info and CONSTANT_Float_info structures represent 4-byte
# numeric (int and float) constants.
#
# The items of these structures are as follows:

class Integer(Constant):
    # u1 tag;
    # u4 bytes;

    def __init__(self, value):
        # The tag item of the CONSTANT_Integer_info structure has the value
        # CONSTANT_Integer (3).
        super().__init__(Constant.CONSTANT_Integer)

        # The bytes item of the CONSTANT_Integer_info structure represents the
        # value of the int constant. The bytes of the value are stored in big-
        # endian (high byte first) order.

        self.value = value

    def __repr__(self):
        return '<Integer %s>' % self.value

    def __eq__(self, other):
        return isinstance(other, Integer) and other.tag == self.tag and other.value == self.value

    def __hash__(self):
        return multihash(self, 'tag', 'value')

    @staticmethod
    def read_info(reader):
        value = reader.read_u4()
        return Integer, (value,)

    def write_info(self, writer):
        writer.write_u4(self.value)


class Float(Constant):
    # u1 tag;
    # u4 bytes;

    def __init__(self, value):
        # The tag item of the CONSTANT_Float_info structure has the value
        # CONSTANT_Float (4).
        super().__init__(Constant.CONSTANT_Float)

        # The bytes item of the CONSTANT_Float_info structure represents the
        # value of the float constant in IEEE 754 floating-point single format
        # (§2.3.2). The bytes of the single format representation are stored in
        # big-endian (high byte first) order.
        #
        # The value represented by the CONSTANT_Float_info structure is
        # determined as follows. The bytes of the value are first converted into
        # an int constant bits. Then:
        #
        #  * If bits is 0x7f800000, the float value will be positive infinity.
        #
        #  * If bits is 0xff800000, the float value will be negative infinity.
        #
        #  * If bits is in the range 0x7f800001 through 0x7fffffff or in the range
        #    0xff800001 through 0xffffffff, the float value will be NaN.
        #
        # In all other cases, let s, e, and m be three values that might be
        # computed from bits:
        #
        #     int s = ((bits >> 31) == 0) ? 1 : -1;
        #     int e = ((bits >> 23) & 0xff);
        #     int m = (e == 0) ?
        #               (bits & 0x7fffff) << 1 :
        #               (bits & 0x7fffff) | 0x800000;
        #
        # Then the float value equals the result of the mathematical expression
        # s · m · 2e-150.

        self.value = value

    def __repr__(self):
        return '<Float %s>' % self.value

    def __eq__(self, other):
        return isinstance(other, Float) and other.tag == self.tag and other.value == self.value

    def __hash__(self):
        return multihash(self, 'tag', 'value')

    @staticmethod
    def read_info(reader):
        value = reader.read_f()
        return Integer, (value,)

    def write_info(self, writer):
        writer.write_f(self.value)


# ------------------------------------------------------------------------
# 4.4.5. The CONSTANT_Long_info and CONSTANT_Double_info Structures
# ------------------------------------------------------------------------

# The CONSTANT_Long_info and CONSTANT_Double_info represent 8-byte numeric (long
# and double) constants.

# All 8-byte constants take up two entries in the constant_pool table of the
# class file. If a CONSTANT_Long_info or CONSTANT_Double_info structure is the
# item in the constant_pool table at index n, then the next usable item in the
# pool is located at index n+2. The constant_pool index n+1 must be valid but is
# considered unusable.

# In retrospect, making 8-byte constants take two constant pool entries was a
# poor choice.

# The items of these structures are as follows:

class Long(Constant):
    # u1 tag;
    # u4 high_bytes;
    # u4 low_bytes;

    # The unsigned high_bytes and low_bytes items of the CONSTANT_Long_info
    # structure together represent the value of the long constant:
    #
    #     ((long) high_bytes << 32) + low_bytes
    #
    # where the bytes of each of high_bytes and low_bytes are stored in big-
    # endian (high byte first) order.

    def __init__(self, value):
        # The tag item of the CONSTANT_Long_info structure has the value
        # CONSTANT_Long (3).
        super().__init__(Constant.CONSTANT_Long)

        self.value = value

    def __repr__(self):
        return '<Long %s>' % self.value

    def __eq__(self, other):
        return isinstance(other, Long) and other.tag == self.tag and other.value == self.value

    def __hash__(self):
        return multihash(self, 'tag', 'value')

    @staticmethod
    def read_info(reader):
        value = reader.read_u8()
        return Long, (value,)

    def write_info(self, writer):
        writer.write_s8(self.value)


class Double(Constant):
    # u1 tag;
    # u4 high_bytes;
    # u4 low_bytes;

    # The high_bytes and low_bytes items of the CONSTANT_Double_info structure
    # together represent the double value in IEEE 754 floating-point double
    # format (§2.3.2). The bytes of each item are stored in big-endian (high
    # byte first) order.
    #
    # The value represented by the CONSTANT_Double_info structure is determined
    # as follows. The high_bytes and low_bytes items are converted into the long
    # constant bits, which is equal to
    #
    # ((long) high_bytes << 32) + low_bytes
    #
    # Then:
    #
    #  * If bits is 0x7ff0000000000000L, the double value will be positive infinity.
    #
    #  * If bits is 0xfff0000000000000L, the double value will be negative infinity.
    #
    #  * If bits is in the range 0x7ff0000000000001L through 0x7fffffffffffffffL or
    #    in the range 0xfff0000000000001L through 0xffffffffffffffffL, the double
    #    value will be NaN.
    #
    # In all other cases, let s, e, and m be three values that might be computed
    # from bits:
    #
    #     int s = ((bits >> 63) == 0) ? 1 : -1;
    #     int e = (int)((bits >> 52) & 0x7ffL);
    #     long m = (e == 0) ?
    #                (bits & 0xfffffffffffffL) << 1 :
    #                (bits & 0xfffffffffffffL) | 0x10000000000000L;
    #
    # Then the floating-point value equals the double value of the mathematical
    # expression s · m · 2e-1075.

    def __init__(self, value):
        # The tag item of the CONSTANT_Double_info structure has the value
        # CONSTANT_Double (6).
        super().__init__(Constant.CONSTANT_Double)

        self.value = value

    def __repr__(self):
        return '<Double %s>' % self.value

    def __eq__(self, other):
        return isinstance(other, Double) and other.tag == self.tag and other.value == self.value

    def __hash__(self):
        return multihash(self, 'tag', 'value')

    @staticmethod
    def read_info(reader):
        value = reader.read_d()
        return Double, (value,)

    def write_info(self, writer):
        writer.write_d(self.value)


# ------------------------------------------------------------------------
# 4.4.6. The CONSTANT_NameAndType_info Structure
# ------------------------------------------------------------------------

# The CONSTANT_NameAndType_info structure is used to represent a field or
# method, without indicating which class or interface type it belongs to:

class NameAndType(Constant):
    # u1 tag;
    # u2 name_index;
    # u2 descriptor_index;

    def __init__(self, name, descriptor):

        # The tag item of the CONSTANT_NameAndType_info structure has the value
        # CONSTANT_NameAndType (12).
        super().__init__(Constant.CONSTANT_NameAndType)

        # The value of the name item must be a valid index into the
        # constant_pool table. The constant_pool entry at that index must be a
        # CONSTANT_Utf8_info (§4.4.7) structure representing either the special
        # method name <init> (§2.9) or a valid unqualified name (§4.2.2)
        # denoting a field or method.

        self.name = Utf8(name)

        # The value of the descriptor item must be a valid index into the
        # constant_pool table. The constant_pool entry at that index must be a
        # CONSTANT_Utf8_info (§4.4.7) structure representing a valid field
        # descriptor (§4.3.2) or method descriptor (§4.3.3).

        self.descriptor = Utf8(descriptor)

    def __repr__(self):
        return '<NameAndType: name:%s descriptor:%s>' % (self.name, self.descriptor)

    def __eq__(self, other):
        return (
            isinstance(other, NameAndType)
            and other.tag == self.tag
            and other.name == self.name
            and other.descriptor == self.descriptor
        )

    def __hash__(self):
        return multihash(self, 'tag', 'name', 'descriptor')

    @staticmethod
    def read_info(reader):
        name = reader.read_u2()
        descriptor = reader.read_u2()
        return NameAndType, (
            read_utf8(name),
            read_utf8(descriptor)
        )

    def write_info(self, writer):
        writer.write_u2(writer.constant_pool.index(self.name))
        writer.write_u2(writer.constant_pool.index(self.descriptor))

    def resolve_info(self, constant_pool):
        self.name.resolve(constant_pool)
        self.descriptor.resolve(constant_pool)


# ------------------------------------------------------------------------
# 4.4.7. The CONSTANT_Utf8_info Structure
# ------------------------------------------------------------------------

# The CONSTANT_Utf8_info structure is used to represent constant string values:

class Utf8(Constant):
    # u1 tag;
    # u2 length;
    # u1 bytes[length];

    def __init__(self, string):

        # The tag item of the CONSTANT_Utf8_info structure has the value
        # CONSTANT_Utf8 (1).
        super().__init__(Constant.CONSTANT_Utf8)

        # The value of the length item gives the number of bytes in the bytes
        # array (not the length of the resulting string). The strings in the
        # CONSTANT_Utf8_info structure are not null-terminated.

        # The bytes array contains the bytes of the string. No byte may have the
        # value (byte)0 or lie in the range (byte)0xf0 - (byte)0xff.

        # String content is encoded in modified UTF-8. Modified UTF-8 strings
        # are encoded so that code point sequences that contain only non-null
        # ASCII characters can be represented using only 1 byte per code point,
        # but all code points in the Unicode codespace can be represented.

        # * Code points in the range '\u0001' to '\u007F' are represented by a
        #   single byte:
        #
        #     Table 4.4.
        #
        #         0   bits 6-0
        #
        #     The 7 bits of data in the byte give the value of the code point represented.
        #
        # * The null code point ('\u0000') and code points in the range '\u0080'
        #   to '\u07FF' are represented by a pair of bytes x and y :
        #
        #     Table 4.5.
        #
        #     x:
        #     Table 4.6.
        #
        #         1   1   0   bits 10-6
        #
        #     y:
        #     Table 4.7.
        #
        #         1   0   bits 5-0
        #
        #     The bytes represent the code point with the value:
        #
        #         ((x & 0x1f) << 6) + (y & 0x3f)
        #
        # * Code points in the range '\u0800' to '\uFFFF' are represented by 3
        #   bytes x, y, and z :
        #
        #     Table 4.8.
        #
        #         x:
        #         Table 4.9.
        #
        #             1   1   1   0   bits 15-12
        #
        #         y:
        #             Table 4.10.
        #
        #             1   0   bits 11-6
        #
        #         z:
        #             Table 4.11.
        #
        #             1   0   bits 5-0
        #
        #     The three bytes represent the code point with the value:
        #
        #         ((x & 0xf) << 12) + ((y & 0x3f) << 6) + (z & 0x3f)
        #
        # * Characters with code points above U+FFFF (so-called supplementary
        #   characters) are represented by separately encoding the two surrogate
        #   code units of their UTF-16 representation. Each of the surrogate
        #   code units is represented by three bytes. This means supplementary
        #   characters are represented by six bytes, u, v, w, x, y, and z :
        #
        #     Table 4.12.
        #
        #         u:
        #             Table 4.13.
        #
        #             1   1   1   0   1   1   0   1
        #
        #         v:
        #             Table 4.14.
        #
        #             1   0   1   0   (bits 20-16)-1
        #
        #         w:
        #             Table 4.15.
        #
        #             1   0   bits 15-10
        #
        #         x:
        #             Table 4.16.
        #
        #             1   1   1   0   1   1   0   1
        #
        #         y:
        #             Table 4.17.
        #
        #             1   0   1   1   bits 9-6
        #
        #         z:
        #             Table 4.18.
        #
        #             1   0   bits 5-0
        #
        #     The six bytes represent the code point with the value:
        #
        #     0x10000 + ((v & 0x0f) << 16) + ((w & 0x3f) << 10) +
        #     ((y & 0x0f) << 6) + (z & 0x3f)
        #
        # The bytes of multibyte characters are stored in the class file in big-
        # endian (high byte first) order.
        #
        # There are two differences between this format and the "standard" UTF-8
        # format. First, the null character (char)0 is encoded using the 2-byte
        # format rather than the 1-byte format, so that modified UTF-8 strings
        # never have embedded nulls. Second, only the 1-byte, 2-byte, and 3-byte
        # formats of standard UTF-8 are used. The Java Virtual Machine does not
        # recognize the four-byte format of standard UTF-8; it uses its own two-
        # times-three-byte format instead.
        #
        # For more information regarding the standard UTF-8 format, see Section
        # 3.9 Unicode Encoding Forms of The Unicode Standard, Version 6.0.0.
        self.value = string
        self._bytes = string.encode('mutf-8')

    def __repr__(self):
        return "b'%s'" % self._bytes.decode('mutf-8')

    def __str__(self):
        return self.value

    @staticmethod
    def read_info(reader):
        length = reader.read_u2()
        string = reader.read_bytes(length).decode('mutf-8')
        return Utf8, (string,)

    def write_info(self, writer):
        writer.write_u2(self.length)
        writer.write_bytes(self.bytes)

    def __eq__(self, other):
        return isinstance(other, Utf8) and other.tag == self.tag and other._bytes == self._bytes

    def __hash__(self):
        return multihash(self, 'tag', '_bytes')

    @property
    def length(self):
        return len(self._bytes)

    @property
    def bytes(self):
        return self._bytes


# ------------------------------------------------------------------------
# 4.4.8. The CONSTANT_MethodHandle_info Structure
# ------------------------------------------------------------------------

# The CONSTANT_MethodHandle_info structure is used to represent a method handle:

# CONSTANT_MethodHandle_info {
#     u1 tag;
#     u1 reference_kind;
#     u2 reference_index;
# }
# The items of the CONSTANT_MethodHandle_info structure are the following:

# tag
# The tag item of the CONSTANT_MethodHandle_info structure has the value
# CONSTANT_MethodHandle (15).

# reference_kind
# The value of the reference_kind item must be in the range 1 to 9. The value
# denotes the kind of this method handle, which characterizes its bytecode
# behavior (§5.4.3.5).

# reference_index
# The value of the reference_index item must be a valid index into the
# constant_pool table.

# If the value of the reference_kind item is 1 (REF_getField), 2
# (REF_getStatic), 3 (REF_putField), or 4 (REF_putStatic), then the
# constant_pool entry at that index must be a CONSTANT_Fieldref_info (§4.4.2)
# structure representing a field for which a method handle is to be created.

# If the value of the reference_kind item is 5 (REF_invokeVirtual), 6
# (REF_invokeStatic), 7 (REF_invokeSpecial), or 8 (REF_newInvokeSpecial), then
# the constant_pool entry at that index must be a CONSTANT_Methodref_info
# structure (§4.4.2) representing a class's method or constructor (§2.9) for
# which a method handle is to be created.

# If the value of the reference_kind item is 9 (REF_invokeInterface), then the
# constant_pool entry at that index must be a CONSTANT_InterfaceMethodref_info
# (§4.4.2) structure representing an interface's method for which a method
# handle is to be created.

# If the value of the reference_kind item is 5 (REF_invokeVirtual), 6
# (REF_invokeStatic), 7 (REF_invokeSpecial), or 9 (REF_invokeInterface), the
# name of the method represented by a CONSTANT_Methodref_info structure must
# not be <init> or <clinit>.

# If the value is 8 (REF_newInvokeSpecial), the name of the method represented
# by a CONSTANT_Methodref_info structure must be <init>.

# ------------------------------------------------------------------------
# 4.4.9. The CONSTANT_MethodType_info Structure
# ------------------------------------------------------------------------

# The CONSTANT_MethodType_info structure is used to represent a method type:

# CONSTANT_MethodType_info {
#     u1 tag;
#     u2 descriptor_index;
# }
# The items of the CONSTANT_MethodType_info structure are as follows:

# tag
# The tag item of the CONSTANT_MethodType_info structure has the value
# CONSTANT_MethodType (16).

# descriptor_index
# The value of the descriptor_index item must be a valid index into the
# constant_pool table. The constant_pool entry at that index must be a
# CONSTANT_Utf8_info (§4.4.7) structure representing a method descriptor
# (§4.3.3).

# ------------------------------------------------------------------------
# 4.4.10. The CONSTANT_InvokeDynamic_info Structure
# ------------------------------------------------------------------------

# The CONSTANT_InvokeDynamic_info structure is used by an invokedynamic
# instruction (§invokedynamic) to specify a bootstrap method, the dynamic
# invocation name, the argument and return types of the call, and optionally, a
# sequence of additional constants called static arguments to the bootstrap
# method.

# CONSTANT_InvokeDynamic_info {
#     u1 tag;
#     u2 bootstrap_method_attr_index;
#     u2 name_and_type_index;
# }
# The items of the CONSTANT_InvokeDynamic_info structure are as follows:

# tag
# The tag item of the CONSTANT_InvokeDynamic_info structure has the value
# CONSTANT_InvokeDynamic (18).

# bootstrap_method_attr_index
# The value of the bootstrap_method_attr_index item must be a valid index into
# the bootstrap_methods array of the bootstrap method table (§4.7.21) of this
# class file.

# name_and_type_index
# The value of the name_and_type_index item must be a valid index into the
# constant_pool table. The constant_pool entry at that index must be a
# CONSTANT_NameAndType_info (§4.4.6) structure representing a method name and
# method descriptor (§4.3.3).



from unittest import expectedFailure

from ..utils import TranspileTestCase


class CollectionsModuleTests(TranspileTestCase):

    #######################################################
    # expected to fail in Python version 3.4 build because
    # it was not documented in Python version <= 3.4
    # # __doc__
    # def test___doc__(self):
        # self.assertCodeExecution("""
            # import collections
            # print(collections.__doc__)
            # """)

    #######################################################
    # __file__
    @expectedFailure
    def test___file__(self):
        self.assertCodeExecution("""
            import collections
            print(collections.__file__)
            """)

    #######################################################
    # __loader__
    @expectedFailure
    def test___loader__(self):
        self.assertCodeExecution("""
            import collections
            print(collections.__loader__)
            """)

    #######################################################
    # __name__
    def test___name__(self):
        self.assertCodeExecution("""
            import collections
            print(collections.__name__)
            """)

    #######################################################
    # __package__
    def test___package__(self):
        self.assertCodeExecution("""
            import collections
            print(collections.__package__)
            """)

    #######################################################
    # __spec__
    @expectedFailure
    def test___spec__(self):
        self.assertCodeExecution("""
            import collections
            print(collections.__spec__)
            """)

    #######################################################
    # namedtuple
    @expectedFailure
    def test_namedtuple(self):
        self.assertCodeExecution("""
            import collections
            print(collections.namedtuple('Point', ['x', 'y']))
            """)

    #######################################################
    # ChainMap
    @expectedFailure
    def test_ChainMap(self):
        self.assertCodeExecution("""
            import collections
            print(collections.ChainMap())
            """)

    #######################################################
    # Counter
    @expectedFailure
    def test_Counter(self):
        self.assertCodeExecution("""
            import collections
            print(collections.Counter())
            """)

    #######################################################
    # UserDict
    @expectedFailure
    def test_UserDict(self):
        self.assertCodeExecution("""
            import collections
            print(collections.UserDict())
            """)

    #######################################################
    # UserList
    @expectedFailure
    def test_UserList(self):
        self.assertCodeExecution("""
            import collections
            print(collections.UserList())
            """)

    #######################################################
    # UserString
    @expectedFailure
    def test_UserString(self):
        self.assertCodeExecution("""
            import collections
            print(collections.UserString("Hello World"))
            """)


class DequeTests(TranspileTestCase):

    pass


class OrderedDictTests(TranspileTestCase):

    def test_creation(self):
        self.assertCodeExecution("""
            import collections
            print(collections.OrderedDict())
            print(collections.OrderedDict({'a': 1}))
            print(collections.OrderedDict(a = 1))
            print(collections.OrderedDict([
                ('a', 1), ('b', 2), ('c', 3), ('d', 4), ('e', 5),
                ('aa', 11), ('bb', 22), ('cc', 33), ('dd', 44), ('ee', 55)]))
            """)

    def test_invalid_argument(self):
        self.assertCodeExecution("""
            import collections

            try:
                od = collections.OrderedDict([1, 2, 3])
                print("should not print this")
            except TypeError as e:
                print(e)

            try:
                od = collections.OrderedDict([(1, 2, 3)])
                print("should not print this")
            except ValueError as e:
                print(e)

            try:
                od = collections.OrderedDict([(1, )])
                print("should not print this")
            except ValueError as e:
                print(e)
            """)

    def test_setitem(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict()
            od['c'] = 3
            print(od)

            od['b'] = 2
            print(od)

            od['a'] = 1
            print(od)

            od['c'] = 0
            print(od)
            """)

    def test_delitem(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            del od['a']
            print(od)

            od['a'] = 0
            print(od)
            """)

    def test_iter(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            # print(iter(od)) Different type prior to Python 3.5
            for i in od:
                print(i)
            for i in iter(od):
                print(i)
            """)

    def test_reversed(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            iter = reversed(od)
            # print(iter) Different type prior to Python 3.5
            for i in iter:
                print(i)
            """)

    def test_clear(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            print(od)
            od.clear()
            print(od)
            """)

    def test_eq(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict([('a', 1), ('b', 2), ('aa', 3)])
            print(od)
            od2 = collections.OrderedDict([('a', 1), ('aa', 3), ('b',2)])
            print(od2)
            print(od == od2)

            d = {'aa': 3, 'a': 1, 'b': 2}
            print(od == d)
            """)

    def test_copy(self):
        self.assertCodeExecution("""
            import collections
            od1 = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            od2 = od1.copy()
            print("od1:", od1)
            print("od2:", od2)
            print(od1 == od2)
            print(od1 is od2)
            od2['e'] = 5
            print("od1:", od1)
            print("od2:", od2)
            print(od1 == od2)
            print(od1['a'] == od2['a'])
            """)

    def test_fromkeys(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict.fromkeys(['c', 'b', 'a'])
            print(od)

            od = collections.OrderedDict.fromkeys(['c', 'b', 'a'], 0)
            print(od)

            try:
                od = collections.OrderedDict.fromkeys(123)
                print("should not print this")
            except TypeError as e:
                print(e)

            try:
                od = collections.OrderedDict([1, 2, 3])
                print("should not print this")
            except TypeError as e:
                print(e)

            try:
                od = collections.OrderedDict([(1, 2, 3)])
                print("should not print this")
            except ValueError as e:
                print(e)

            try:
                od = collections.OrderedDict([(1, )])
                print("should not print this")
            except ValueError as e:
                print(e)
            """)

    def test_items(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            items = od.items()

            # print(items) Different type prior to Python 3.5
            print(list(items))
            print(len(items))
            for i, j in items:
                print(i, j)

            od['aa'] = 11
            print(od)
            # print(items) Different type prior to Python 3.5
            print(list(items))
            """)

    def test_items_reversed(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            try:
                items = reversed(od.items())

                print(items)
                for i, j in items:
                    print(i, j)

                od['aa'] = 11
                print(od)
                print(items)
            except TypeError as e:
                # not implemented in Python version < 3.5
                print(e)
            """)

    def test_keys(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            keys = od.keys()

            # print(keys) Different type prior to Python 3.5
            print(list(keys))
            print(len(keys))
            for i in keys:
                print(i)

            od['aa'] = 11
            print(od)
            # print(keys) Different type prior to Python 3.5
            print(list(keys))
            """)

    def test_keys_reversed(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            try:
                keys = reversed(od.keys())

                print(keys)
                for i in keys:
                    print(i)

                od['aa'] = 11
                print(od)
                print(keys)
            except TypeError as e:
                # not implemented in Python version < 3.5
                print(e)
            """)

    def test_values(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            values = od.values()

            # print(values) Different type prior to Python 3.5
            print(list(values))
            print(len(values))
            for i in values:
                print(i)

            od['aa'] = 11
            print(od)
            # print(values) Different type prior to Python 3.5
            print(list(values))
            """)

    def test_values_reversed(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            try:
                values = reversed(od.values())

                print(values)
                for i in values:
                    print(i)

                od['aa'] = 11
                print(od)
                print(values)
            except TypeError as e:
                # not implemented in Python version < 3.5
                print(e)
            """)

    def test_pop(self):
        self.assertCodeExecution("""
            import collections

            od = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            print(od.pop('c'))
            print(od)
            print(od.pop('d', 4))
            print(od)

            try:
                print(od.pop('d'))
                print("should not print this")
            except KeyError as e:
                print(e)
            """)

    def test_popitem(self):
        self.assertCodeExecution("""
            import collections

            od = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            print(od.popitem())
            print(od)
            print(od.popitem(last=False))
            print(od)
            print(od.popitem(last=True))
            print(od)

            try:
                print(od.popitem())
                print("should not print this")
            except KeyError as e:
                print(e)
            """)

    def test_update(self):
        self.assertCodeExecution("""
            import collections

            od = collections.OrderedDict([('c', 3), ('b', 2), ('a', 1)])
            od.update()
            print(od)
            od.update({'a': 0})
            print(od)
            od.update({'d': 4})
            print(od)
            od.update(e = 5)
            print(od)
            od.update(collections.OrderedDict([('f', 6), ('g', 7), ('h', 8)]))
            print(od)

            try:
                od.update([(1, 2, 3)])
                print("should not print this")
            except ValueError as e:
                print(e)

            try:
                od.update([(1, )])
                print("should not print this")
            except ValueError as e:
                print(e)

            try:
                od.update("abc")
                print("should not print this")
            except ValueError as e:
                print(e)
            """)

    def test_move_to_end(self):
        self.assertCodeExecution("""
            import collections
            od = collections.OrderedDict.fromkeys('abcde')
            od.move_to_end('b')
            print(''.join(od.keys()))
            od.move_to_end('b', last=False)
            print(''.join(od.keys()))
            """)


class DefaultDictTests(TranspileTestCase):

    def test_creation(self):
        self.assertCodeExecution("""
            import collections
            print(collections.defaultdict())
            print(collections.defaultdict(int))
            print(collections.defaultdict(list, {'a': 1}))
            """)

    def test_invalid_first_argument(self):
        self.assertCodeExecution("""
            import collections
            try:
                d = collections.defaultdict(123)
            except TypeError as e:
                print(e)
            """)

    def test_dict_method(self):
        self.assertCodeExecution("""
            import collections
            d = collections.defaultdict(int)
            print(d.get("key"))
            print(d.setdefault("default"))
            print(d)
            """)

    def test_default_list(self):
        self.assertCodeExecution("""
            from collections import defaultdict

            s = [('yellow', 1), ('blue', 2), ('yellow', 3), ('blue', 4), ('red', 1)]
            d = defaultdict(list)
            for k, v in s:
                d[k].append(v)
            print(sorted(d.items()))
            """)

    def test_default_int(self):
        self.assertCodeExecution("""
            from collections import defaultdict

            s = 'mississippi'
            d = defaultdict(int)
            for k in s:
                d[k] += 1
            print(sorted(d.items()))
            """)

    def test_default_callable(self):
        self.assertCodeExecution("""
            from collections import defaultdict

            def constant_factory(value):
                return lambda: value
            d = defaultdict(constant_factory('<missing>'))
            d.update(name='John', action='ran')
            print('%(name)s %(action)s to %(object)s' % d)
            """)

    def test_default_set(self):
        self.assertCodeExecution("""
            from collections import defaultdict

            s = [('red', 1), ('blue', 2), ('red', 3), ('blue', 4), ('red', 1), ('blue', 4)]
            d = defaultdict(set)
            for k, v in s:
                d[k].add(v)
            print(sorted(d.items()))
            """)



from .. utils import TranspileTestCase, BuiltinTwoargFunctionTestCase


class MapTests(TranspileTestCase):
    base_code = """
        iterable = %s
        def testish(x):
            return %s
        it = map(testish, iterable)
        print(it)
        print(next(it))
        print(next(it))
        print(next(it))
        try:
            print(next(it))
        except StopIteration:
            pass
    """

    def test_bool(self):
        self.assertCodeExecution(self.base_code % ("[True, False, True]", "bool(x)"))

    def test_bytearray(self):
        self.assertCodeExecution(self.base_code % ("b'123'", "x"))

    def test_float(self):
        self.assertCodeExecution(self.base_code % ("[3.14, 2.17, 1.0]", "x > 1"))

    def test_int(self):
        self.assertCodeExecution(self.base_code % ("[1, 2, 3]", "x * 2"))


class BuiltinMapFunctionTests(BuiltinTwoargFunctionTestCase, TranspileTestCase):
    functions = ["map"]

    not_implemented = [
        'test_class_obj'
    ]



import os
import sys
from unittest import expectedFailure

from ..utils import TranspileTestCase


class TimeModuleTests(TranspileTestCase):

    #######################################################
    # _STRUCT_TM_ITEMS
    @expectedFailure
    def test__STRUCT_TM_ITEMS(self):
        self.assertCodeExecution("""
            import time
            print(time._STRUCT_TM_ITEMS)
            """)

    #######################################################
    # __doc__
    def test___doc__(self):
        if sys.hexversion > 0x03060400:
            # Docstring was truncated in Python 3.6.4
            substitutions = {
                '': [
                    "\n" +
                    "Variables:\n" +
                    "\n" +
                    "timezone -- difference in seconds between UTC and local standard time\n" +
                    "altzone -- difference in  seconds between UTC and local DST time\n" +
                    "daylight -- whether local time should reflect DST\n" +
                    "tzname -- tuple of (standard time zone name, DST time zone name)\n" +
                    "\n" +
                    "Functions:\n" +
                    "\n" +
                    "time() -- return current time in seconds since the Epoch as a float\n" +
                    "clock() -- return CPU time since process start as a float\n" +
                    "sleep() -- delay for a number of seconds given as a float\n" +
                    "gmtime() -- convert seconds since Epoch to UTC tuple\n" +
                    "localtime() -- convert seconds since Epoch to local time tuple\n" +
                    "asctime() -- convert time tuple to string\n" +
                    "ctime() -- convert time in seconds to string\n" +
                    "mktime() -- convert local time tuple to seconds since Epoch\n" +
                    "strftime() -- convert time tuple to string according to format specification\n" +
                    "strptime() -- parse string to time tuple according to format specification\n" +
                    "tzset() -- change the local timezone"
                ]
            }
        else:
            substitutions = None

        self.assertCodeExecution(
            """
            import time
            print(time.__doc__)
            """,
            substitutions=substitutions
        )

    #######################################################
    # __file__
    @expectedFailure
    def test___file__(self):
        self.assertCodeExecution("""
            import time
            print(time.__file__)
            """)

    #######################################################
    # __loader__
    @expectedFailure
    def test___loader__(self):
        self.assertCodeExecution("""
            import time
            print(time.__loader__)
            """)

    #######################################################
    # __name__
    def test___name__(self):
        self.assertCodeExecution("""
            import time
            print(time.__name__)
            """)

    #######################################################
    # __package__
    def test___package__(self):
        self.assertCodeExecution("""
            import time
            print(time.__package__)
            """)

    #######################################################
    # __spec__
    @expectedFailure
    def test___spec__(self):
        self.assertCodeExecution("""
            import time
            print(time.__spec__)
            """)

    #######################################################
    # altzone
    @expectedFailure
    def test_altzone(self):
        self.assertCodeExecution("""
            import time
            print(time.altzone)
            """)

    #######################################################
    # asctime
    @expectedFailure
    def test_asctime(self):
        self.assertCodeExecution("""
            import time
            print(time.asctime())
            """)

    #######################################################
    # clock
    def test_clock(self):
        # Since we can't know exactly what CPU time will be used,
        # and CPU time will vary between implementations,
        # this test validates that clock returns a float < 0.01s
        sleepy_time = 1
        diff_offset = sleepy_time if os.name == 'nt' else 0
        # On Windows, time.clock includes the time spent in time.sleep
        # however on Unix it does not.
        self.assertCodeExecution("""
            import time
            start = time.clock()
            time.sleep({sleepy_time})
            end = time.clock()
            diff = end - start - {diff_offset}
            print(type(diff))
            print(diff < 0.1)
            """.format(sleepy_time=sleepy_time, diff_offset=diff_offset))

    #######################################################
    # ctime
    def test_ctime(self):
        self.assertCodeExecution("""
            import time
            print(time.ctime()[:10], time.ctime()[-4:])
            """)

    def test_ctime_with_parameter(self):
        self.assertCodeExecution("""
            import time
            print(time.ctime(0))
            print(time.ctime(1000))
            now = time.time()
            print(time.ctime((now - (now % 3600))))
            print(time.ctime(1000.67))
            try:
                time.ctime('today')
            except Exception as e:
                print(e)
            try:
                time.ctime([1,2])
            except Exception as e:
                print(e)
            try:
                time.ctime((1,2))
            except Exception as e:
                print(e)
            time.ctime(None)
        """)

    #######################################################
    # daylight
    @expectedFailure
    def test_daylight(self):
        self.assertCodeExecution("""
            import time
            print(time.daylight)
            """)

    #######################################################
    # get_clock_info
    @expectedFailure
    def test_get_clock_info(self):
        self.assertCodeExecution("""
            import time
            print(time.get_clock_info())
            """)

    #######################################################
    # gmtime
    @expectedFailure
    def test_gmtime(self):
        self.assertCodeExecution("""
            import time
            print(time.gmtime())
            """)

    #######################################################
    # localtime
    @expectedFailure
    def test_localtime(self):
        self.assertCodeExecution("""
            import time
            print(time.localtime())
            """)

    #######################################################
    # mktime
    @expectedFailure
    def test_mktime(self):
        self.assertCodeExecution("""
            import time
            print(time.mktime())
            """)

    #######################################################
    # monotonic
    @expectedFailure
    def test_monotonic(self):
        self.assertCodeExecution("""
            import time
            print(time.monotonic())
            """)

    #######################################################
    # perf_counter
    @expectedFailure
    def test_perf_counter(self):
        self.assertCodeExecution("""
            import time
            print(time.perf_counter())
            """)

    #######################################################
    # process_time
    @expectedFailure
    def test_process_time(self):
        self.assertCodeExecution("""
            import time
            print(time.process_time())
            """)

    #######################################################
    # sleep
    def test_sleep(self):
        self.assertCodeExecution("""
            import time
            print(time.sleep(1))
            """)

    #######################################################
    # strftime
    @expectedFailure
    def test_strftime(self):
        self.assertCodeExecution("""
            import time
            print(time.strftime())
            """)

    #######################################################
    # strptime
    @expectedFailure
    def test_strptime(self):
        self.assertCodeExecution("""
            import time
            print(time.strptime())
            """)

    #######################################################
    # struct_time
    @expectedFailure
    def test_struct_time(self):
        self.assertCodeExecution("""
            import time
            print(time.struct_time())
            """)

    #######################################################
    # time
    def test_time(self):
        self.assertCodeExecution("""
            import time
            print(int(time.time() / 10000))
            """)

    #######################################################
    # timezone
    @expectedFailure
    def test_timezone(self):
        self.assertCodeExecution("""
            import time
            print(time.timezone)
            """)

    #######################################################
    # tzname

    @expectedFailure
    def test_tzname(self):
        self.assertCodeExecution("""
            import time
            print(time.tzname)
            """)

    #######################################################
    # tzset
    @expectedFailure
    def test_tzset(self):
        self.assertCodeExecution("""
            import time
            print(time.tzset())
            """)



from unittest import expectedFailure

from ..utils import TranspileTestCase


class IteratorTests(TranspileTestCase):
    def test_iterable(self):
        self.assertCodeExecution("""
            class TestSet:
                def __init__(self, values):
                    self.values = values
                    self.i = -1

                def __iter__(self):
                    return self

                def __next__(self):
                    try:
                        self.i += 1
                        if self.values[self.i] % 2 == 0:
                            return 'even'
                        else:
                            return self.values[self.i]
                    except IndexError:
                        raise StopIteration()

            s = TestSet([1, 2, 3, 4, 5])
            for val in s:
                print(val)

            """)

    def test_generator_as_iterable(self):
        self.assertCodeExecution("""
            class Interview:
                def __init__(self, start, stop):
                    self.start = start
                    self.stop = stop

                def __iter__(self):
                    for i in range(self.start, self.stop):
                        found = False
                        if i % 2 == 0:
                            yield 'fizz'
                            found = True
                        if i % 3 == 0:
                            yield 'buzz'
                            found = True
                        if not found:
                            yield i

            myinterview = Interview(1, 20)

            for i in myinterview:
                print(i)
            """)

    @expectedFailure
    def test_bad_iter(self):
        self.assertCodeExecution("""
            try:
                x = iter()
            except TypeError as err:
                print(err)
        """)



import sys
import argparse
import voc

from .klass import Class


def dump(filename):
    with open(filename, 'rb') as infile:
        Class.read(infile, debug=sys.stdout)


def main():
    parser = argparse.ArgumentParser(
        prog='vod',
        description='Debugging tool to decompile class files.'
    )

    parser.add_argument(
        '--version', '-v',
        action='version',
        version='voc %s' % voc.__version__,
    )

    parser.add_argument(
        'file',
    )

    args = parser.parse_args()
    dump(vars(args)['file'])


if __name__ == "__main__":
    main()



from ..utils import TranspileTestCase


class ReturnTests(TranspileTestCase):
    def test_return_methods(self):
        # Empty return
        self.assertCodeExecution("""
            def m():
                print('before return')
                return
                print('after return')

            print(m())
        """)

        # Single return
        self.assertCodeExecution("""
            def m():
                print('before return')
                return 1
                print('after return')

            print(m() + 1)
        """)

        # Multiple return
        self.assertCodeExecution("""
            def m():
                print('before return')
                return 1, 2
                print('after return')

            print(m())
        """)



import os

from ..utils import TranspileTestCase


class ImportTests(TranspileTestCase):
    def test_import_stdlib_module(self):
        "You can import a Python module implemented in Java (a native stdlib shim)"
        self.assertCodeExecution(
            """
            import time

            time.time()

            print("Done.")
            """)

    def test_import_module(self):
        "You can import a Python module implemented in Python"
        self.assertCodeExecution(
            """
            import example

            example.some_method()

            print("Done.")
            """,
            extra_code={
                'example':
                    """
                    print("Now we're in the example module")

                    def some_method():
                        print("Now we're calling a module method")
                    """
            })

    def test_import_module_main(self):
        "You can import a Python module with if __name__ == '__main__'"
        self.assertCodeExecution(
            """
            import example

            print('A')

            if __name__ == "__main__":
                print('main')

            print('B')
            """,
            extra_code={
                'example':
                    """
                    print('C')

                    if __name__ == "__main__":
                        print('example main')

                    print('D')
                    """
            })

    def test_multiple_module_import(self):
        "You can import a multiple Python modules implemented in Python"
        self.assertCodeExecution(
            """
            import example, other

            example.some_method()

            other.other_method()

            print("Done.")
            """,
            extra_code={
                'example':
                    """
                    print("Now we're in the example module")

                    def some_method():
                        print("Now we're calling a module method")
                    """,
                'other':
                    """
                    print("Now we're in the other module")

                    def other_method():
                        print("Now we're calling another module method")
                    """
            })

    def test_full_dotted_path(self):
        self.assertCodeExecution(
            """
            import example.submodule

            example.submodule.some_method()

            print("Done.")
            """,
            extra_code={
                'example.__init__':
                    """
                    print("Initializing the example module")
                    """,
                'example.submodule':
                    """
                    print("Now we're in example.submodule")

                    def some_method():
                        print("Now we're calling a submodule method")
                    """
            })

    def test_module_from_dotted_path(self):
        self.assertCodeExecution(
            """
            from example import submodule

            submodule.some_method()

            print("Done.")
            """,
            extra_code={
                'example.__init__':
                    """
                    print("Initializing the example module")
                    """,
                'example.submodule':
                    """
                    print("Now we're in example.submodule")

                    def some_method():
                        print("Now we're calling a submodule method")
                    """
            })

    def test_symbol_from_dotted_path(self):
        self.assertCodeExecution(
            """
            from example.submodule import some_method

            some_method()

            print("Done.")
            """,
            extra_code={
                'example.__init__':
                    """
                    print("Initializing the example module")
                    """,
                'example.submodule':
                    """
                    print("Now we're in example.submodule")

                    def some_method():
                        print("Now we're calling a submodule method")
                    """
            })

    def test_ambiguous_symbol_from_dotted_path(self):
        self.assertCodeExecution(
            """
            from example.thing import Thing

            Thing()

            print("Done.")
            """,
            extra_code={
                'example.__init__':
                    """
                    print("Initializing the example module")
                    """,
                'example.thing':
                    """
                    print("Now we're in example.thing")

                    class Thing:
                        def __init__(self):
                            print("Now we're creating a Thing")

                    """
            })

    def test_full_deep_dotted_path(self):
        self.assertCodeExecution(
            """
            import example.submodule.subsubmodule.another

            example.submodule.subsubmodule.another.another_method()

            print("Done.")
            """,
            extra_code={
                'example.__init__':
                    """
                    print("Initializing the example module")
                    """,
                'example.submodule.__init__':
                    """
                    print("Now we're in example.submodule.__init__")
                    """,
                'example.submodule.other':
                    """
                    print("Now we're in example.submodule.other")

                    def other_method():
                        print("Now we're calling a submodule method")
                    """,
                'example.submodule.subsubmodule.__init__':
                    """
                    print("Now we're in example.submodule.subsubmodule.__init__")
                    """,
                'example.submodule.subsubmodule.another':
                    """
                    print("Now we're in example.submodule.subsubmodule.another")

                    def another_method():
                        print("Now we're calling a subsubmodule method")
                    """
            })

    def test_module_from_deep_dotted_path(self):
        self.assertCodeExecution(
            """
            from example.submodule.subsubmodule import another

            another.another_method()

            print("Done.")
            """,
            extra_code={
                'example.__init__':
                    """
                    print("Initializing the example module")
                    """,
                'example.submodule.__init__':
                    """
                    print("Now we're in example.submodule.__init__")
                    """,
                'example.submodule.other':
                    """
                    print("Now we're in example.submodule.other")

                    def other_method():
                        print("Now we're calling a submodule method")
                    """,
                'example.submodule.subsubmodule.__init__':
                    """
                    print("Now we're in example.submodule.subsubmodule.__init__")
                    """,
                'example.submodule.subsubmodule.another':
                    """
                    print("Now we're in example.submodule.subsubmodule.another")

                    def another_method():
                        print("Now we're calling a subsubmodule method")
                    """
            })

    def test_symbol_from_deep_dotted_path(self):
        self.assertCodeExecution(
            """
            from example.submodule.subsubmodule.another import another_method

            another_method()

            print("Done.")
            """,
            extra_code={
                'example.__init__':
                    """
                    print("Initializing the example module")
                    """,
                'example.submodule.__init__':
                    """
                    print("Now we're in example.submodule.__init__")
                    """,
                'example.submodule.other':
                    """
                    print("Now we're in example.submodule.other")

                    def other_method():
                        print("Now we're calling a submodule method")
                    """,
                'example.submodule.subsubmodule.__init__':
                    """
                    print("Now we're in example.submodule.subsubmodule.__init__")
                    """,
                'example.submodule.subsubmodule.another':
                    """
                    print("Now we're in example.submodule.subsubmodule.another")

                    def another_method():
                        print("Now we're calling a subsubmodule method")
                    """
            })

    def test_symbol_import(self):
        self.assertCodeExecution(
            """
            from example import some_method

            some_method()

            print("Done.")
            """,
            extra_code={
                'example':
                    """
                    print("Now we're in the example module")

                    def some_method():
                        print("Now we're calling a module method")
                    """
            })

    def test_multiple_symbol_import(self):
        self.assertCodeExecution(
            """
            from example import some_method, other_method

            print("Call some method...")
            some_method()

            print("Call another method...")
            other_method()

            try:
                print("But this will fail...")
                third_method()
            except NameError:
                print("Which it does.")

            print("Done.")
            """,
            extra_code={
                'example':
                    """
                    print("Now we're in the example module")

                    def some_method():
                        print("Now we're calling a module method")

                    def other_method():
                        print("Now we're calling another module method")

                    def third_method():
                        print("This shouldn't be called")

                    """
            })

    def test_import_star(self):
        self.assertCodeExecution(
            """
            from example import *

            print("Call some method...")
            some_method()

            print("Call another method...")
            other_method()

            print("Call a third method...")
            third_method()

            print("Done.")
            """,
            extra_code={
                'example':
                    """
                    print("Now we're in the example module")

                    def some_method():
                        print("Now we're calling a module method")

                    def other_method():
                        print("Now we're calling another module method")

                    def third_method():
                        print("Now we're calling a third module method")

                    """
            }, run_in_function=False)

    def test_import_star_with_all(self):
        self.assertCodeExecution(
            """
            from example import *

            print("Call some method...")
            some_method()

            print("Call another method...")
            other_method()

            try:
                print("But this will fail...")
                third_method()
            except NameError:
                print("Which it does.")

            print("Done.")
            """,
            extra_code={
                'example':
                    """
                    __all__ = ['some_method', 'other_method']

                    print("Now we're in the example module")

                    def some_method():
                        print("Now we're calling a module method")

                    def other_method():
                        print("Now we're calling another module method")

                    def third_method():
                        print("This shouldn't be called")

                    """
            }, run_in_function=False)

    def test_import_from_dot(self):
        self.assertCodeExecution(
            """
            from example import submodule2

            submodule2.method()

            print("Done.")
            """,
            extra_code={
                'example.__init__':
                    """
                    """,
                'example.submodule1':
                    """
                    def method():
                        print("Calling method in submodule1")
                    """,
                'example.submodule2':
                    """
                    from . import submodule1

                    def method():
                        print("Calling method in submodule2")
                        submodule1.method()
                    """,
            })

    def test_import_from_local_dot(self):
        self.assertCodeExecution(
            """
            from submodule import method1, method2

            method1()
            method2()

            print("Done.")
            """,
            extra_code={
                'submodule.__init__':
                    """
                    print("in submodule/__init__.py")
                    from .modulea import method2

                    def method1():
                        print("Calling method in submodule.__init__")
                    """,
                'submodule.modulea':
                    """
                    print("in submodule/modulea.py")

                    def method2():
                        print("Calling method in submodule.modulea")
                    """,
            })

    def test_import_from_local_dot_sibling(self):
        self.assertCodeExecution(
            """
            from submodule.moduleb import method3, method2

            method3()
            method2()

            print("Done.")
            """,
            extra_code={
                'submodule.__init__':
                    """
                    print("in submodule/__init__.py")

                    def method1():
                        print("Calling method1 in submodule.__init__")
                    """,
                'submodule.modulea':
                    """
                    print("in submodule/modulea.py")

                    def method2():
                        print("Calling method2 in submodule.modulea")
                    """,
                'submodule.moduleb':
                    """
                    print("in submodule/moduleb.py")
                    from .modulea import method2

                    def method3():
                        print("Calling method3 in submodule.moduleb")
                    """,
            })

    def test_import_from_local_dot_deep(self):
        self.assertCodeExecution(
            """
            from submodule import method1, method2, method3, method4

            method1()
            method2()
            method3()
            method4()

            print("Done.")
            """,
            extra_code={
                'submodule.__init__':
                    """
                    print("in submodule/__init__.py")
                    from .modulea import method2
                    from .subsubmodule import method3, method4

                    def method1():
                        print("Calling method1 in submodule.__init__")
                    """,
                'submodule.modulea':
                    """
                    print("in submodule/modulea.py")

                    def method2():
                        print("Calling method2 in submodule.modulea")
                    """,
                'submodule.subsubmodule.__init__':
                    """
                    print("in submodule/subsubmodule/__init__.py")
                    from .submodulea import method4

                    def method3():
                        print("Calling method3 in submodule.subsubmodule.__init__")
                    """,
                'submodule.subsubmodule.submodulea':
                    """
                    print("in submodule/subsubmodule/submodulea.py")

                    def method4():
                        print("Calling method4 in submodule.subsubmodule.submodula")
                    """,
            })

    def test_import_from_deep_upstream(self):
        self.assertCodeExecution(
            """
            from submodule.subsubmodule.submodulea import method

            method()

            print("Done.")
            """,
            extra_code={
                'submodule.__init__':
                    """
                    print("in submodule/__init__.py")

                    def method1():
                        print("Calling method in submodule.__init__")
                    """,
                'submodule.modulea':
                    """
                    print("in submodule/modulea.py")

                    def method2():
                        print("Calling method in submodule.modulea")
                    """,
                'submodule.moduleb':
                    """
                    print("in submodule/moduleb.py")

                    def method3():
                        print("Calling method in submodule.moduleb")
                    """,
                'submodule.modulec':
                    """
                    print("in submodule/modulec.py")

                    def method4():
                        print("Calling method in submodule.modulec")
                    """,
                'submodule.moduled.__init__':
                    """
                    print("in submodule/moduled/__init__.py")

                    def method5():
                        print("Calling method in submodule.moduled")
                    """,
                'submodule.moduled.submoduled':
                    """
                    print("in submodule/moduled/submoduled.py")

                    def method6():
                        print("Calling method in submodule.moduled.submoduled")
                    """,
                'submodule.subsubmodule.__init__':
                    """
                    print("in submodule/subsubmodule/__init__.py")

                    def method7():
                        print("Calling method in submodule.subsubmodule.__init__")
                    """,
                'submodule.subsubmodule.submodulea':
                    """
                    print("in submodule/subsubmodule/submodulea.py")
                    from .. import moduleb
                    from ..modulec import method4
                    from ..moduled import method5, submoduled

                    def method():
                        print("Calling method4 in submodule.subsubmodule.submodulea")
                        moduleb.method3()
                        method4()
                        method5()
                        submoduled.method6()
                    """,
            })


class NativeImportTests(TranspileTestCase):
    def test_import_java_module_static_method(self):
        "You can invoke a static method from a native Java namespace"
        self.assertJavaExecution(
            """
            from java import lang

            props = lang.System.getProperties()
            print(props.get("file.separator"))

            print("Done.")
            """,
            """
            %s
            Done.
            """ % os.path.sep)

    def test_import_java_class_static_method(self):
        "You can invoke a static method from a native Java class"
        self.assertJavaExecution(
            """
            from java.lang import System

            props = System.getProperties()
            print(props.get("file.separator"))

            print("Done.")
            """,
            """
            %s
            Done.
            """ % os.path.sep)

    def test_import_java_module(self):
        "You can import a native Java namespace as a Python module"
        self.assertJavaExecution(
            """
            from java import lang

            buf = lang.StringBuilder()
            buf.append('Hello, ')
            buf.append('World')
            print(buf.toString())

            print("Done.")
            """,
            """
            Hello, World
            Done.
            """)

    def test_import_java_class(self):
        "You can import a native Java class as a Python module"
        self.assertJavaExecution(
            """
            from java.lang import StringBuilder

            buf = StringBuilder()
            buf.append('Hello, ')
            buf.append('World')
            print(buf.toString())

            print("Done.")
            """,
            """
            Hello, World
            Done.
            """)


class BuiltinsImportTests(TranspileTestCase):
    def test_import_builtins(self):
        self.assertCodeExecution("""
            import builtins
            print(builtins.abs(-42))
            print("Done")
            """)

    def test_import_from_builtins(self):
        self.assertCodeExecution("""
            from builtins import abs
            print(abs(-42))
            print("Done")
            """)

    def test_import_from_builtins_as(self):
        self.assertCodeExecution("""
            from builtins import abs as _abs
            print(_abs(-42))
            print("Done")
            """)
